from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import xarray as xr

import xclim.indices as xci
from xclim import atmos, convert, core, set_options
from xclim.core.calendar import build_climatology_bounds, percentile_doy
from xclim.core.units import convert_units_to

K2C = 273.15


class TestRainOnFrozenGround:
    @pytest.mark.parametrize("chunks", [{"time": 366}, None])
    def test_3d_data_with_nans(self, open_dataset, chunks):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

        pr = ds.pr.copy()
        pr.values[1, 10] = np.nan

        if chunks:
            ds = ds.chunk(chunks)

        out = atmos.rain_on_frozen_ground_days(pr, ds.tas, freq="YS")
        np.testing.assert_array_equal(out.sel(location="Montréal"), [np.nan, 4, 5, 3])


class TestRainSeason:
    # @pytest.mark.parametrize("chunks", [{"time": 366}, None])
    def test_3d_data_with_nans(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

        pr = ds.pr.isel(location=0).copy()
        pr[{"time": [0, 10, 100]}] = np.nan
        out = {}
        out["start"], out["end"], out["length"] = atmos.rain_season(
            pr,
            freq="YS-JAN",
            window_dry_end=5,
            date_min_start="01-01",
            date_min_end="01-01",
        )
        out_arr = np.array([out[var].values for var in ["start", "end", "length"]])
        out_exp = np.array(
            [
                [np.nan, 12.0, 6.0, 27.0],
                [np.nan, np.nan, 141.0, np.nan],
                [np.nan, 354.0, 135.0, 339.0],
            ]
        )
        np.testing.assert_array_equal(out_arr, out_exp)


class TestPrecipAccumulation:
    nc_pr = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"
    nc_tasmin = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_pr).pr  # mm/s
        prMM = open_dataset(self.nc_pr).pr
        prMM *= 86400
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.precip_accumulation(pr, freq="MS")
        out2 = atmos.precip_accumulation(prMM, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.precip_accumulation(pr, freq="MS")

        np.testing.assert_allclose(out1, out2, rtol=1e-7, atol=1e-4)
        np.testing.assert_allclose(out1, out3, rtol=1e-7, atol=1e-4)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        pr_tot = x1.sum()

        np.testing.assert_almost_equal(pr_tot, out1.values[0, 0, 0], 4)

        assert np.isnan(out1.values[0, 1, 0])
        assert np.isnan(out1.values[0, -1, -1])

    def test_with_different_phases(self, open_dataset):
        # test with different phases
        pr = open_dataset(self.nc_pr).pr  # mm/s
        tasmin = open_dataset(self.nc_tasmin).tasmin  # K

        out_tot = atmos.precip_accumulation(pr, freq="MS")
        out_sol = atmos.solid_precip_accumulation(pr, tas=tasmin, freq="MS")
        out_liq = atmos.liquid_precip_accumulation(pr, tas=tasmin, freq="MS")

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)

        assert "solid" in out_sol.description
        assert "liquid" in out_liq.description
        assert out_sol.standard_name == "lwe_thickness_of_snowfall_amount"

        # With a non-default threshold
        out_sol = atmos.solid_precip_accumulation(pr, tas=tasmin, thresh="40 degF", freq="MS")
        out_liq = atmos.liquid_precip_accumulation(pr, tas=tasmin, thresh="40 degF", freq="MS")

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)


class TestPrecipAverage:
    nc_pr = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"
    nc_tasmin = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_pr).pr  # mm/s
        prMM = open_dataset(self.nc_pr).pr
        prMM *= 86400
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.precip_average(pr, freq="MS")
        out2 = atmos.precip_average(prMM, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.precip_average(pr, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 5)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        pr_mean = x1.mean()

        np.testing.assert_almost_equal(pr_mean, out1.values[0, 0, 0], 4)

        assert np.isnan(out1.values[0, 1, 0])

        assert np.isnan(out1.values[0, -1, -1])

    def test_with_different_phases(self, open_dataset):
        # test with different phases
        pr = open_dataset(self.nc_pr).pr  # mm/s
        tasmin = open_dataset(self.nc_tasmin).tasmin  # K

        out_tot = atmos.precip_average(pr, freq="MS")
        out_sol = atmos.solid_precip_average(pr, tas=tasmin, freq="MS")
        out_liq = atmos.liquid_precip_average(pr, tas=tasmin, freq="MS")

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)

        assert "solid" in out_sol.description
        assert "liquid" in out_liq.description
        assert out_sol.standard_name == "lwe_average_of_snowfall_amount"

        # With a non-default threshold
        out_sol = atmos.solid_precip_average(pr, tas=tasmin, thresh="40 degF", freq="MS")
        out_liq = atmos.liquid_precip_average(pr, tas=tasmin, thresh="40 degF", freq="MS")

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)


class TestStandardizedPrecip:
    nc_ds = "sdba/CanESM2_1950-2100.nc"

    @pytest.mark.slow
    def test_3d_data_with_nans(self, open_dataset):
        # test with data
        ds = open_dataset(self.nc_ds)
        pr = ds.pr.sel(time=slice("2000"))  # kg m-2 s-1
        prMM = convert_units_to(pr, "mm/day", context="hydro")  # noqa
        # put a nan somewhere
        prMM.values[10] = np.nan
        pr.values[10] = np.nan

        out1 = atmos.standardized_precipitation_index(
            pr, freq="MS", window=1, dist="gamma", method="APP", fitkwargs={"floc": 0}
        )
        out2 = atmos.standardized_precipitation_index(
            prMM, freq="MS", window=1, dist="gamma", method="APP", fitkwargs={"floc": 0}
        )
        np.testing.assert_array_almost_equal(out1, out2, 3)

        # preparing water_budget for SPEI test
        with xr.set_options(keep_attrs=True):
            tasmax = ds.tasmax
            tas = tasmax - 2.5
            tasmin = tasmax - 5
            wb = xci.water_budget(pr, None, tasmin, tasmax, tas, None)
            wbMM = convert_units_to(wb, "mm/day", context="hydro")  # noqa

        out3 = atmos.standardized_precipitation_evapotranspiration_index(
            wb,
            freq="MS",
            window=1,
            dist="gamma",
            method="APP",
            fitkwargs={"floc": 0},
        )
        out4 = atmos.standardized_precipitation_evapotranspiration_index(
            wbMM,
            freq="MS",
            window=1,
            dist="gamma",
            method="APP",
            fitkwargs={"floc": 0},
        )

        np.testing.assert_array_almost_equal(out3, out4, 3)


class TestWetDays:
    nc_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = "5 mm/d"
        out1 = atmos.wetdays(pr, thresh=pr_min, freq="MS")
        out2 = atmos.wetdays(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.wetdays(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        wd1 = (x1 >= int(pr_min.split(" ")[0])).sum()

        assert wd1 == out1.values[0, 0, 0]

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(wds.values[0, -1, -1]))


class TestWetPrcptot:
    """Testing of prcptot with wet days"""

    def test_simple(self, atmosds):
        pr = atmosds.pr

        thresh = "1 mm/day"
        out = atmos.wet_precip_accumulation(pr, thresh=thresh)

        # Reference value
        t = core.units.convert_units_to(thresh, pr, context="hydro")
        pa = atmos.precip_accumulation(pr.where(pr >= t, 0))
        np.testing.assert_array_equal(out, pa)


class TestDailyIntensity:
    # testing of wet_day and daily_pr_intensity, both are related
    nc_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        # compute with both skipna options
        pr_min = "2 mm/d"
        # dis = daily_pr_intensity(pr, pr_min=pr_min, freq='MS', skipna=True)

        out1 = atmos.daily_pr_intensity(pr, thresh=pr_min, freq="MS")
        out2 = atmos.daily_pr_intensity(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.daily_pr_intensity(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values

        di1 = x1[x1 >= int(pr_min.split(" ")[0])].mean()
        # buffer = np.ma.masked_invalid(x2)
        # di2 = buffer[buffer >= pr_min].mean()

        assert np.allclose(di1, out1.values[0, 0, 0])
        # assert (np.allclose(di1, dis.values[0, 0, 0]))
        assert np.isnan(out1.values[0, 1, 0])
        # assert (np.allclose(di2, dis.values[0, 1, 0]))
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(dis.values[0, -1, -1]))


class TestMaxPrIntensity:
    def test_simple(self, pr_hr_series):
        pr1 = pr_hr_series(np.zeros(366 * 24))
        pr1[10:20] += np.arange(10)
        pr2 = pr_hr_series(np.ones(366 * 24))

        pr = xr.concat([pr1, pr2], dim="site")
        out = atmos.max_pr_intensity(pr, window=2, freq="YS")
        np.testing.assert_array_almost_equal(out.isel(time=0), [8.5 * 3600, 3600])

    def test_indexing(self, pr_hr_series):
        pr = pr_hr_series(np.zeros(366 * 24))
        pr[20:30] += np.arange(10)

        out1 = atmos.max_pr_intensity(pr, window=2, freq="YS")
        out2 = atmos.max_pr_intensity(pr, window=2, freq="YS", doy_bounds=[200, 1])
        np.testing.assert_array_almost_equal(out1.isel(time=0), [8.5 * 3600])
        np.testing.assert_array_almost_equal(out2.isel(time=0), [2.5 * 3600])


class TestMax1Day:
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.max_1day_precipitation_amount(pr, freq="MS")
        out2 = atmos.max_1day_precipitation_amount(prMM, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.max_1day_precipitation_amount(pr, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values
        rx1 = x1.max()

        assert np.allclose(rx1, out1.values[0, 0, 0])
        assert np.isnan(out1.values[0, 1, 0])
        assert np.isnan(out1.values[0, -1, -1])


class TestMaxNDay:
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    @pytest.mark.parametrize(
        "units,factor,chunks",
        [
            ("mm/day", 86400.0, None),
            ("kg m-2 s-1", 1, None),
            ("mm/s", 1, {"time": 73.0}),
        ],
    )
    def test_3d_data_with_nans(self, open_dataset, units, factor, chunks):
        # test with 3d data
        pr1 = open_dataset(self.nc_file).pr
        pr2 = open_dataset(self.nc_file, chunks=chunks).pr
        pr2.values *= factor
        pr2.attrs["units"] = units
        # put a nan somewhere
        pr2.values[10, 1, 0] = np.nan
        pr1.values[10, 1, 0] = np.nan
        wind = 3
        out1 = atmos.max_n_day_precipitation_amount(pr1, window=wind, freq="MS")
        out2 = atmos.max_n_day_precipitation_amount(pr2, window=wind, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)

        x1 = pr1[:31, 0, 0].values * 86400
        df = pd.DataFrame({"pr": x1})
        rx3 = df.rolling(wind).sum().max()

        assert np.allclose(rx3, out1.values[0, 0, 0])
        assert np.isnan(out1.values[0, 1, 0])
        assert np.isnan(out1.values[0, -1, -1])

    def test_indexing(self, open_dataset):
        pr = open_dataset(self.nc_file).pr
        out1 = atmos.max_n_day_precipitation_amount(pr, window=7, freq="YS")
        out2 = atmos.max_n_day_precipitation_amount(pr, window=7, freq="YS", month=[6, 7, 8])

        np.testing.assert_allclose(out1.values[0, 0, 0], 55.48, atol=1e-2)
        np.testing.assert_allclose(out2.values[0, 0, 0], 50.57, atol=1e-2)


class TestMaxConsecWetDays:
    nc_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = "5 mm/d"
        out1 = atmos.maximum_consecutive_wet_days(pr, thresh=pr_min, freq="MS")
        out2 = atmos.maximum_consecutive_wet_days(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.maximum_consecutive_wet_days(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0] * 0.0
        x1[5:10] = 10
        x1.attrs["units"] = "mm/day"
        cwd1 = atmos.maximum_consecutive_wet_days(x1, freq="MS")

        assert cwd1 == 5

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])


class TestMaxConsecDryDays:
    nc_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = "5 mm/d"
        out1 = atmos.maximum_consecutive_dry_days(pr, thresh=pr_min, freq="MS")
        out2 = atmos.maximum_consecutive_dry_days(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.maximum_consecutive_dry_days(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0] * 0.0 + 50.0
        x1[5:10] = 0
        x1.attrs["units"] = "mm/day"
        cdd1 = atmos.maximum_consecutive_dry_days(x1, freq="MS")

        assert cdd1 == 5

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])


class TestSnowfallDate:
    tasmin_file = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"
    pr_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    @classmethod
    def get_snowfall(cls, nimbus):
        dnr = xr.merge(
            (
                xr.open_dataset(nimbus.fetch(cls.pr_file), engine="h5netcdf"),
                xr.open_dataset(nimbus.fetch(cls.tasmin_file), engine="h5netcdf"),
            )
        )
        return convert.snowfall_approximation(dnr.pr, tas=dnr.tasmin, thresh="-0.5 degC", method="binary")

    def test_first_snowfall(self, nimbus):
        with set_options(check_missing="skip"):
            fs = atmos.first_snowfall(prsn=self.get_snowfall(nimbus), thresh="0.5 mm/day")

        np.testing.assert_array_equal(
            fs[:, [0, 45, 82], [10, 105, 155]],
            np.array(
                [
                    [[1, 1, 1], [1, 1, 1], [11, np.nan, np.nan]],
                    [[254, 256, 277], [274, 292, 275], [300, np.nan, np.nan]],
                ]
            ),
        )

    def test_last_snowfall(self, nimbus):
        with set_options(check_missing="skip"):
            ls = atmos.last_snowfall(prsn=self.get_snowfall(nimbus), thresh="0.5 mm/day")

        np.testing.assert_array_equal(
            ls[:, [0, 45, 82], [10, 105, 155]],
            np.array(
                [
                    [[155, 151, 129], [127, 157, 110], [106, np.nan, np.nan]],
                    [[365, 363, 363], [365.0, 364, 364], [362, np.nan, np.nan]],
                ]
            ),
        )


class TestDaysWithSnow:
    def test_simple(self, open_dataset, prsn_series):
        prsn = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").prsn
        out = atmos.days_with_snow(prsn, low="0 kg m-2 s-1", high="1e6 kg m-2 s-1")
        np.testing.assert_array_equal(out[1], [np.nan, 162, 159, 126, np.nan])


def test_days_over_precip_doy_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = percentile_doy(pr, window=5, per=80)

    out1 = atmos.days_over_precip_doy_thresh(pr, per)
    np.testing.assert_array_equal(out1[1, :, 0], np.array([81, 60, 68, 78]))

    out2 = atmos.days_over_precip_doy_thresh(pr, per, thresh="2 mm/d")
    np.testing.assert_array_equal(out2[1, :, 0], np.array([80, 59, 66, 78]))

    assert "only days with at least 2 mm/d are counted." in out2.description
    assert "[80]th percentile" in out2.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out2.attrs["description"]
    assert "5 day(s)" in out2.attrs["description"]


def test_days_over_precip_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = pr.quantile(0.8, "time", keep_attrs=True)
    per.attrs["climatology_bounds"] = build_climatology_bounds(pr)

    out = atmos.days_over_precip_thresh(pr, per)

    np.testing.assert_allclose(out[1, :], np.array([80.0, 64.0, 65.0, 83.0]), atol=0.001)
    assert "80.0th percentile" in out.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out.attrs["description"]


def test_days_over_precip_thresh__seasonal_indexer(open_dataset):
    # GIVEN
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = pr.quantile(0.8, "time", keep_attrs=True)
    # WHEN
    out = atmos.days_over_precip_thresh(pr, per, freq="YS", date_bounds=("01-10", "12-31"))
    # THEN
    np.testing.assert_almost_equal(out[0], np.array([81.0, 66.0, 66.0, 75.0]))


def test_fraction_over_precip_doy_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = percentile_doy(pr, window=5, per=80)

    out = atmos.fraction_over_precip_doy_thresh(pr, per)
    np.testing.assert_allclose(out[1, :, 0], np.array([0.803, 0.747, 0.745, 0.806]), atol=0.001)

    out = atmos.fraction_over_precip_doy_thresh(pr, per, thresh="0.002 m/d")
    np.testing.assert_allclose(out[1, :, 0], np.array([0.822, 0.780, 0.771, 0.829]), atol=0.001)

    assert "only days with at least 0.002 m/d are included" in out.description
    assert "[80]th percentile" in out.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out.attrs["description"]
    assert "5 day(s)" in out.attrs["description"]


def test_fraction_over_precip_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = pr.quantile(0.8, "time", keep_attrs=True)
    per.attrs["climatology_bounds"] = build_climatology_bounds(pr)

    out = atmos.fraction_over_precip_thresh(pr, per)

    np.testing.assert_allclose(out[1, :], np.array([0.839, 0.812, 0.776, 0.864]), atol=0.001)

    assert "80.0th percentile" in out.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out.attrs["description"]


def test_liquid_precip_ratio(open_dataset):
    ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

    out = atmos.liquid_precip_ratio(pr=ds.pr, tas=ds.tas, thresh="0 degC", freq="YS")
    np.testing.assert_allclose(out[:, 0], np.array([0.919, 0.805, 0.525, 0.740, 0.993]), atol=1e3)

    with set_options(cf_compliance="raise"):
        # Test if tasmax is allowed
        out = atmos.liquid_precip_ratio(pr=ds.pr, tas=ds.tasmax, thresh="33 degF", freq="YS")
        np.testing.assert_allclose(out[:, 0], np.array([0.975, 0.921, 0.547, 0.794, 0.999]), atol=1e3)
        assert "where temperature is above 33 degf." in out.description


def test_dry_spell(atmosds):
    pr = atmosds.pr

    events = atmos.dry_spell_frequency(pr, thresh="3 mm", window=7, freq="YS")
    total_d_sum = atmos.dry_spell_total_length(pr, thresh="3 mm", window=7, op="sum", freq="YS")
    total_d_max = atmos.dry_spell_total_length(pr, thresh="3 mm", window=7, op="max", freq="YS")
    max_d_sum = atmos.dry_spell_max_length(pr, thresh="3 mm", window=7, op="sum", freq="YS")
    max_d_max = atmos.dry_spell_max_length(pr, thresh="3 mm", window=7, op="max", freq="YS")
    total_d_sum = total_d_sum.sel(location="Halifax", drop=True).isel(time=slice(0, 2))
    total_d_max = total_d_max.sel(location="Halifax", drop=True).isel(time=slice(0, 2))
    max_d_sum = max_d_sum.sel(location="Halifax", drop=True).isel(time=slice(0, 2))
    max_d_max = max_d_max.sel(location="Halifax", drop=True).isel(time=slice(0, 2))

    np.testing.assert_allclose(events[0:2, 0], [5, 7], rtol=1e-1)
    np.testing.assert_allclose(total_d_sum, [50, 53], rtol=1e-1)
    np.testing.assert_allclose(total_d_max, [68, 97], rtol=1e-1)
    np.testing.assert_allclose(max_d_sum, [14, 10], rtol=1e-1)
    np.testing.assert_allclose(max_d_max, [14, 14], rtol=1e-1)

    assert (
        "The annual number of dry periods of 7 day(s) or more, "
        "during which the total precipitation on a window of 7 day(s) is below 3 mm."
    ) in events.description
    assert "The annual number of days in dry periods of 7 day(s) or more" in total_d_sum.description
    assert "The annual number of days in dry periods of 7 day(s) or more" in total_d_max.description
    assert "The maximum annual number of consecutive days in a dry period of 7 day(s) or more" in max_d_sum.description
    assert "The maximum annual number of consecutive days in a dry period of 7 day(s) or more" in max_d_max.description


def test_dry_spell_total_length_indexer(pr_series):
    pr = pr_series([np.nan] + [1] * 4 + [0] * 10 + [1] * 350, start="1900-01-01", units="mm/d")
    out = atmos.dry_spell_total_length(
        pr,
        window=7,
        op="sum",
        thresh="3.1 mm",
        freq="MS",
    )
    np.testing.assert_allclose(out, [np.nan] + [0] * 11)

    out = atmos.dry_spell_total_length(
        pr, window=7, op="sum", thresh="3.1 mm", freq="MS", date_bounds=("01-10", "12-31")
    )
    np.testing.assert_allclose(out, [9] + [0] * 11)


def test_dry_spell_max_length_indexer(pr_series):
    pr = pr_series([np.nan] + [1] * 4 + [0] * 10 + [1] * 350, start="1900-01-01", units="mm/d")
    out = atmos.dry_spell_max_length(
        pr,
        window=7,
        op="sum",
        thresh="3.1 mm",
        freq="MS",
    )
    np.testing.assert_allclose(out, [np.nan] + [0] * 11)

    out = atmos.dry_spell_total_length(
        pr, window=7, op="sum", thresh="3.1 mm", freq="MS", date_bounds=("01-10", "12-31")
    )
    np.testing.assert_allclose(out, [9] + [0] * 11)


def test_dry_spell_frequency_op(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    test_sum = atmos.dry_spell_frequency(pr, thresh="3 mm", window=7, freq="MS", op="sum")
    test_max = atmos.dry_spell_frequency(pr, thresh="3 mm", window=7, freq="MS", op="max")

    np.testing.assert_allclose(test_sum[0, :14], [1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0], rtol=1e-1)
    np.testing.assert_allclose(test_max[0, :14], [1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 2, 1], rtol=1e-1)
    assert (
        "The monthly number of dry periods of 7 day(s) or more, "
        "during which the total precipitation on a window of 7 day(s) is below 3 mm."
    ) in test_sum.description
    assert (
        "The monthly number of dry periods of 7 day(s) or more, "
        "during which the maximal precipitation on a window of 7 day(s) is below 3 mm."
    ) in test_max.description


class TestSnowfallMeteoSwiss:
    tasmin_file = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"
    pr_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    @classmethod
    def get_snowfall(cls, nimbus):
        dnr = xr.merge(
            (
                xr.open_dataset(nimbus.fetch(cls.pr_file), engine="h5netcdf"),
                xr.open_dataset(nimbus.fetch(cls.tasmin_file), engine="h5netcdf"),
            )
        )
        return convert.snowfall_approximation(dnr.pr, tas=dnr.tasmin, thresh="-0.5 degC", method="binary")

    def test_snowfall_frequency(self, nimbus):
        prsn = self.get_snowfall(nimbus)
        with set_options(check_missing="skip"):
            sf = atmos.snowfall_frequency(prsn=prsn, thresh="1 mm/day")
        expected = np.array(
            [
                [
                    [27.624, 29.834, 25.414],
                    [22.652, 25.414, 22.652],
                    [12.155, 0.0, 0.0],
                ],
                [
                    [23.913, 23.370, 20.652],
                    [17.391, 15.761, 13.043],
                    [4.891, 0.0, 0.0],
                ],
            ]
        )
        np.testing.assert_allclose(
            sf[:, [0, 45, 82], [10, 105, 155]],
            expected,
            rtol=1e-3,
        )

    def test_snowfall_intensity(self, nimbus):
        prsn = self.get_snowfall(nimbus)
        with set_options(check_missing="skip"):
            si = atmos.snowfall_intensity(prsn=prsn, thresh="1 mm/day")
        expected = np.array(
            [
                [
                    [3.585, 3.839, 4.446],
                    [5.148, 5.884, 5.764],
                    [5.910, 0.0, 0.0],
                ],
                [
                    [5.093, 5.284, 5.539],
                    [5.946, 6.840, 9.702],
                    [9.522, 0.0, 0.0],
                ],
            ]
        )
        np.testing.assert_allclose(
            si[:, [0, 45, 82], [10, 105, 155]],
            expected,
            rtol=1e-3,
        )



"""
Run-Length Algorithms Submodule
===============================

Computation of statistics on runs of True values in boolean arrays.
"""

from __future__ import annotations

from collections import namedtuple
from collections.abc import Callable, Sequence
from datetime import datetime
from typing import Literal
from warnings import warn

import numpy as np
import pandas as pd
import xarray as xr
from numba import njit

from xclim.core import DateStr, DayOfYearStr
from xclim.core.options import OPTIONS, RUN_LENGTH_UFUNC
from xclim.core.utils import lazy_indexing, uses_dask
from xclim.indices.helpers import resample_map

npts_opt = 9000
"""
Arrays with less than this number of data points per slice will trigger
the use of the ufunc version of run lengths algorithms.
"""


def use_ufunc(
    ufunc_1dim: bool | str,
    da: xr.DataArray,
    dim: str = "time",
    freq: str | None = None,
    index: str = "first",
) -> bool:
    """
    Return whether the ufunc version of run length algorithms should be used with this DataArray or not.

    If ufunc_1dim is 'from_context', the parameter is read from xclim's global (or context) options.
    If it is 'auto', this returns False for dask-backed array and for arrays with more than :py:const:`npts_opt`
    points per slice along `dim`.

    Parameters
    ----------
    ufunc_1dim : {'from_context', 'auto', True, False}
        The method for handling the ufunc parameters.
    da : xr.DataArray
        Input array.
    dim : str
        The dimension along which to find runs.
    freq : str
        Resampling frequency.
    index : {'first', 'last'}
        If 'first' (default), the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    bool
        If ufunc_1dim is "auto", returns True if the array is on dask or too large.
        Otherwise, returns ufunc_1dim.
    """
    if ufunc_1dim is True and freq is not None:
        raise ValueError("Resampling after run length operations is not implemented for 1d method")

    if ufunc_1dim == "from_context":
        ufunc_1dim = OPTIONS[RUN_LENGTH_UFUNC]

    if ufunc_1dim == "auto":
        ufunc_1dim = not uses_dask(da) and (da.size // da[dim].size) < npts_opt
    # If resampling after run length is set up for the computation, the 1d method is not implemented
    # Unless ufunc_1dim is specifically set to False (in which case we flag an error above),
    # we simply forbid this possibility.
    return (index == "first") and ufunc_1dim and (freq is None)


def _is_chunked(da, dim):
    """Check if `da` has non-trivial chunks"""
    chunksize = (da.chunksizes.get(dim, (-1,)))[0]
    return chunksize not in (-1, da[dim].size)


def resample_and_rl(
    da: xr.DataArray,
    resample_before_rl: bool,
    compute: Callable,
    *args,
    freq: str,
    dim: str = "time",
    **kwargs,
) -> xr.DataArray:
    r"""
    Wrap run length algorithms to control if resampling occurs before or after the algorithms.

    Parameters
    ----------
    da : xr.DataArray
        N-dimensional array (boolean).
    resample_before_rl : bool
        Determines whether if input arrays of runs `da` should be separated in period before
        or after the run length algorithms are applied.
    compute : Callable
        Run length function to apply.
    *args : Any
        Positional arguments needed in `compute`.
    freq : str
        Resampling frequency.
    dim : str
        The dimension along which to find runs.
    **kwargs : Any
        Keyword arguments needed in `compute`.

    Returns
    -------
    xr.DataArray
        Output of compute resampled according to frequency {freq}.
    """
    if resample_before_rl:
        out = resample_map(
            da,
            dim,
            freq,
            compute,
            map_kwargs={"args": args, "freq": None, "dim": dim, **kwargs},
        )
    else:
        out = compute(da, *args, dim=dim, freq=freq, **kwargs)
    return out


def _smallest_uint(da, dim):
    for dtype in [np.uint8, np.uint16, np.uint32, np.uint64]:
        if np.iinfo(dtype).max > da[dim].size:
            return dtype
    return np.uint64


# Specifying `one` allows in-place multiplication *=
@njit
def _cumsum_reset_np(arr, index, one):
    """100110111 -> 100120123"""
    # run the cumsum and prod backwards or forward
    it = range(1, arr.shape[-1]) if index == "last" else range(arr.shape[-1] - 2, -1, -1)
    for i in it:
        # this works because we assume to have 1's and 0's
        arr[..., i] *= arr[..., i - it.step] + one
    return arr


def _cumsum_reset_xr(da, dim, index, reset_on_zero):
    """100110111 -> 100120123"""
    # `index="first"` case: Algorithm is applied on inverted array and output is inverted back
    if index == "first":
        da = da[{dim: slice(None, None, -1)}]

    # Example: da == 100110111 -> cs_s == 100120123
    cs = da.cumsum(dim=dim)  # cumulative sum  e.g. 111233456
    cond = da == 0 if reset_on_zero else da.isnull()  # reset condition
    cs2 = cs.where(cond)  # keep only numbers at positions of zeroes e.g. N11NN3NNN (default)
    cs2[{dim: 0}] = 0  # put a zero in front e.g. 011NN3NNN
    cs2 = cs2.ffill(dim=dim)  # e.g. 011113333
    out = cs - cs2
    if index == "first":
        out = out[{dim: slice(None, None, -1)}]
    return out


def _cumsum_reset(
    da: xr.DataArray,
    dim: str = "time",
    index: str = "last",
) -> xr.DataArray:
    """
    Compute the cumulative sum for each series of numbers separated by zero.

    Parameters
    ----------
    da : xr.DataArray
        Input array.
    dim : str
        Dimension name along which the cumulative sum is taken.
    index : {'first', 'last'}
        If 'first', the largest value of the cumulative sum is indexed with the first element in the run.
        If 'last'(default), with the last element in the run.

    Returns
    -------
    xr.DataArray
        An array with cumulative sums.

    Notes
    -----
    This function converts `nan` to 0, and uses integers. The fast track will only work correctly
    with binary entries. To avoid these limitations, use directly `_cumsum_reset_xr`.
    """
    # Most of our indicators implicitly transform nans into False, e.g. rle(da0 > thresh)
    # da0 might contain NaNs, these will be converted to False. If the input `da` received
    # here contains NaNs, we adopt the same behaviour.
    typ = _smallest_uint(da, dim)

    # fast track
    if not _is_chunked(da, dim):
        # only int can be used in this case
        da = da.fillna(typ(0))
        out = xr.apply_ufunc(
            _cumsum_reset_np,
            da,
            input_core_dims=[[dim]],
            output_core_dims=[[dim]],
            dask="parallelized",
            kwargs={"index": index, "one": typ(1)},
        )
    else:
        out = _cumsum_reset_xr(da, dim, index, reset_on_zero=True)
    return out


# TODO: Check if rle would be more performant with ffill/bfill instead of two times [{dim: slice(None, None, -1)}]
def rle(
    da: xr.DataArray,
    dim: str = "time",
    index: str = "first",
) -> xr.DataArray:
    """
    Run length.

    Despite its name, this is not an actual run-length encoder : it returns an array of the same shape
    as the input with 0 where the input was <= 0, nan where the input was > 0, except on the first (or last) element
    of each run of consecutive > 0 values, where it is set to the sum of the elements within the run.
    For an actual run length encoder, see :py:func:`rle_1d`.

    Usually, the input would be a boolean mask and the first element of each run would then be set to
    the run's length (thus the name), but the function also accepts int and float inputs.

    Parameters
    ----------
    da : xr.DataArray
        Input array.
    dim : str
        Dimension name.
    index : {'first', 'last'}
        If 'first' (default), the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    xr.DataArray
        The run length array.
    """
    # "first" case: Algorithm is applied on inverted array and output is inverted back
    if index == "first":
        da = da[{dim: slice(None, None, -1)}]

    # Get cumulative sum for each series of 1, e.g. da == 100110111 -> cs_s == 100120123
    cs_s = _cumsum_reset(da, dim)

    # Keep total length of each series (and also keep 0's), e.g. 100120123 -> 100N20NN3
    # Keep numbers with a 0 to the right and also the last number
    # We could keep a -1 instead of nan to save space? Like this we could keep int types.
    cs_s = cs_s.where(da.shift({dim: -1}, fill_value=0) == 0)
    out = cs_s.where(da > 0, 0)  # Reinsert 0's at their original place

    # Inverting back if needed e.g. 100N20NN3 -> 3NN02N001. This is the output of
    # `rle` for 111011001 with index == "first"
    if index == "first":
        out = out[{dim: slice(None, None, -1)}]

    return out


def rle_statistics(
    da: xr.DataArray,
    reducer: str,
    window: int,
    dim: str = "time",
    freq: str | None = None,
    ufunc_1dim: str | bool = "from_context",
    index: str = "first",
) -> xr.DataArray:
    """
    Return the length of consecutive run of True values, according to a reducing operator.

    Parameters
    ----------
    da : xr.DataArray
        N-dimensional array (boolean).
    reducer : str
        Name of the reducing function.
    window : int
        Minimal length of consecutive runs to be included in the statistics.
    dim : str
        Dimension along which to calculate consecutive run; Default: 'time'.
    freq : str
        Resampling frequency.
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points.  Using 1D_ufunc=True is typically more efficient
        for DataArray with a small number of grid points.
        It can be modified globally through the "run_length_ufunc" global option.
    index : {'first', 'last'}
        If 'first' (default), the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    xr.DataArray, [int]
        Length of runs of True values along dimension, according to the reducing function (float)
        If there are no runs (but the data is valid), returns 0.
    """
    ufunc_1dim = use_ufunc(ufunc_1dim, da, dim=dim, index=index, freq=freq)
    if ufunc_1dim:
        rl_stat = statistics_run_ufunc(da, reducer, window, dim)
    else:
        d = rle(da, dim=dim, index=index)

        def get_rl_stat(d):
            rl_stat = getattr(d.where(d >= window), reducer)(dim=dim)
            rl_stat = xr.where((d.isnull() | (d < window)).all(dim=dim), 0, rl_stat)
            return rl_stat

        if freq is None:
            rl_stat = get_rl_stat(d)
        else:
            rl_stat = resample_map(d, dim, freq, get_rl_stat)
    return rl_stat


def longest_run(
    da: xr.DataArray,
    dim: str = "time",
    freq: str | None = None,
    ufunc_1dim: str | bool = "from_context",
    index: str = "first",
) -> xr.DataArray:
    """
    Return the length of the longest consecutive run of True values.

    Parameters
    ----------
    da : xr.DataArray
        N-dimensional array (boolean).
    dim : str
        Dimension along which to calculate consecutive run; Default: 'time'.
    freq : str
        Resampling frequency.
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points.  Using 1D_ufunc=True is typically more efficient
        for DataArray with a small number of grid points.
        It can be modified globally through the "run_length_ufunc" global option.
    index : {'first', 'last'}
        If 'first', the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    xr.DataArray, [int]
        Length of the longest run of True values along dimension (int).
    """
    return rle_statistics(
        da,
        reducer="max",
        window=1,
        dim=dim,
        freq=freq,
        ufunc_1dim=ufunc_1dim,
        index=index,
    )


def windowed_run_events(
    da: xr.DataArray,
    window: int,
    dim: str = "time",
    freq: str | None = None,
    ufunc_1dim: str | bool = "from_context",
    index: str = "first",
) -> xr.DataArray:
    """
    Return the number of runs of a minimum length.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum run length.
        When equal to 1, an optimized version of the algorithm is used.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    freq : str
        Resampling frequency.
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points.  Using 1D_ufunc=True is typically more efficient
        for DataArray with a small number of grid points.
        Ignored when `window=1`. It can be modified globally through the "run_length_ufunc" global option.
    index : {'first', 'last'}
        If 'first', the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    xr.DataArray, [int]
        Number of distinct runs of a minimum length (int).
    """
    ufunc_1dim = use_ufunc(ufunc_1dim, da, dim=dim, index=index, freq=freq)

    if ufunc_1dim:
        out = windowed_run_events_ufunc(da, window, dim)

    else:
        if window == 1:
            shift = 1 * (index == "first") + -1 * (index == "last")
            d = xr.where(da.shift({dim: shift}, fill_value=0) == 0, 1, 0)
            d = d.where(da == 1, 0)
        else:
            d = rle(da, dim=dim, index=index)
            d = xr.where(d >= window, 1, 0)
        if freq is not None:
            d = d.resample({dim: freq})
        out = d.sum(dim=dim)

    return out


def windowed_run_count(
    da: xr.DataArray,
    window: int,
    dim: str = "time",
    freq: str | None = None,
    ufunc_1dim: str | bool = "from_context",
    index: str = "first",
) -> xr.DataArray:
    """
    Return the number of consecutive true values in array for runs at least as long as given duration.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum run length.
        When equal to 1, an optimized version of the algorithm is used.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    freq : str
        Resampling frequency.
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points. Using 1D_ufunc=True is typically more efficient
        for DataArray with a small number of grid points.
        Ignored when `window=1`. It can be modified globally through the "run_length_ufunc" global option.
    index : {'first', 'last'}
        If 'first', the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    xr.DataArray, [int]
        Total number of `True` values part of a consecutive runs of at least `window` long.
    """
    ufunc_1dim = use_ufunc(ufunc_1dim, da, dim=dim, index=index, freq=freq)

    if ufunc_1dim:
        out = windowed_run_count_ufunc(da, window, dim)

    elif window == 1 and freq is None:
        out = da.sum(dim=dim)

    else:
        d = rle(da, dim=dim, index=index)
        d = d.where(d >= window, 0)
        if freq is not None:
            d = d.resample({dim: freq})
        out = d.sum(dim=dim)

    return out


def windowed_max_run_sum(
    da: xr.DataArray,
    window: int,
    dim: str = "time",
    freq: str | None = None,
    index: str = "first",
) -> xr.DataArray:
    """
    Return the maximum sum of consecutive float values for runs at least as long as the given window length.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray.
    window : int
        Minimum run length.
        When equal to 1, an optimized version of the algorithm is used.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    freq : str
        Resampling frequency.
    index : {'first', 'last'}
        If 'first', the run length is indexed with the first element in the run.
        If 'last', with the last element in the run.

    Returns
    -------
    xr.DataArray, [float]
        Cumulative sum of input values part of a consecutive runs of at least `window` long.

    Notes
    -----
    The input `da` is expected to be non-negative, e.g. temperature exceedances
    `(tasmax - thresh).clip(0,None)`
    """
    if window == 1 and freq is None:
        # using _cumsum_reset_xr instead of rle to be able to handle a float
        out = _cumsum_reset_xr(da, dim=dim, index=index, reset_on_zero=True).max(dim=dim)

    else:
        # using _cumsum_reset_xr instead of rle to be able to handle a float
        d_rse = _cumsum_reset_xr(da, dim=dim, index=index, reset_on_zero=True)
        d_rle = rle(da > 0, dim=dim, index=index)

        d = d_rse.where(d_rle >= window, 0)
        if freq is not None:
            d = d.resample({dim: freq})
        out = d.max(dim=dim)

    return out


def _boundary_run(
    da: xr.DataArray,
    window: int,
    dim: str,
    freq: str | None,
    coord: str | bool | None,
    ufunc_1dim: str | bool,
    position: str,
) -> xr.DataArray:
    """
    Return the index of the first item of the first or last run of at least a given length.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
        When equal to 1, an optimized version of the algorithm is used.
    dim : str
        Dimension along which to calculate consecutive run.
    freq : str
        Resampling frequency.
    coord : Optional[str]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points.  Using 1D_ufunc=True is typically more efficient
        for DataArray with a small number of grid points.
        Ignored when `window=1`. It can be modified globally through the "run_length_ufunc" global option.
    position : {"first", "last"}
        Determines if the algorithm finds the "first" or "last" run

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of first item in first (last) valid run.
        Returns np.nan if there are no valid runs.
    """

    # FIXME: The logic here should not use outside scope variables, but rather pass them as arguments.
    def coord_transform(out, da):
        """Transforms indexes to coordinates if needed, and drops obsolete dim."""
        if coord:
            crd = da[dim]
            if isinstance(coord, str):
                crd = getattr(crd.dt, coord)
            out = lazy_indexing(crd, out)

        if dim in out.coords:
            out = out.drop_vars(dim)
        return out

    # FIXME: The logic here should not use outside scope variables, but rather pass them as arguments.
    # general method to get indices (or coords) of first run
    def find_boundary_run(runs, position):
        if position == "last":
            runs = runs[{dim: slice(None, None, -1)}]
        dmax_ind = runs.argmax(dim=dim)
        # If there are no runs, dmax_ind will be 0: We must replace this with NaN
        out = dmax_ind.where(dmax_ind != runs.argmin(dim=dim))
        if position == "last":
            out = runs[dim].size - out - 1
            runs = runs[{dim: slice(None, None, -1)}]
        out = coord_transform(out, runs)
        return out

    ufunc_1dim = use_ufunc(ufunc_1dim, da, dim=dim, freq=freq)

    da = da.fillna(0)  # We expect a boolean array, but there could be NaNs nonetheless
    if window == 1:
        if freq is not None:
            out = resample_map(da, dim, freq, find_boundary_run, map_kwargs={"position": position})
        else:
            out = find_boundary_run(da, position)

    elif ufunc_1dim:
        if position == "last":
            da = da[{dim: slice(None, None, -1)}]
        out = first_run_ufunc(x=da, window=window, dim=dim)
        if position == "last" and not coord:
            out = da[dim].size - out - 1
            da = da[{dim: slice(None, None, -1)}]
        out = coord_transform(out, da)

    else:
        # _cusum_reset is an intermediate step in rle, which is sufficient here
        d = _cumsum_reset(da, dim=dim, index=position)
        d = xr.where(d >= window, 1, 0)
        # for "first" run, return "first" element in the run (and conversely for "last" run)
        if freq is not None:
            out = resample_map(d, dim, freq, find_boundary_run, map_kwargs={"position": position})
        else:
            out = find_boundary_run(d, position)

    return out


def first_run(
    da: xr.DataArray,
    window: int,
    dim: str = "time",
    freq: str | None = None,
    coord: str | bool | None = False,
    ufunc_1dim: str | bool = "from_context",
) -> xr.DataArray:
    """
    Return the index of the first item of the first run of at least a given length.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
        When equal to 1, an optimized version of the algorithm is used.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    freq : str
        Resampling frequency.
    coord : Optional[str]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points.  Using 1D_ufunc=True is typically more efficient
        for DataArray with a small number of grid points.
        Ignored when `window=1`. It can be modified globally through the "run_length_ufunc" global option.

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of first item in first valid run.
        Returns np.nan if there are no valid runs.
    """
    out = _boundary_run(
        da,
        window=window,
        dim=dim,
        freq=freq,
        coord=coord,
        ufunc_1dim=ufunc_1dim,
        position="first",
    )
    return out


def last_run(
    da: xr.DataArray,
    window: int,
    dim: str = "time",
    freq: str | None = None,
    coord: str | bool | None = False,
    ufunc_1dim: str | bool = "from_context",
) -> xr.DataArray:
    """
    Return the index of the last item of the last run of at least a given length.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
        When equal to 1, an optimized version of the algorithm is used.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    freq : str
        Resampling frequency.
    coord : Optional[str]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').
    ufunc_1dim : Union[str, bool]
        Use the 1d 'ufunc' version of this function : default (auto) will attempt to select optimal
        usage based on number of data points.  Using `1D_ufunc=True` is typically more efficient
        for a DataArray with a small number of grid points.
        Ignored when `window=1`. It can be modified globally through the "run_length_ufunc" global option.

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of last item in last valid run.
        Returns np.nan if there are no valid runs.
    """
    out = _boundary_run(
        da,
        window=window,
        dim=dim,
        freq=freq,
        coord=coord,
        ufunc_1dim=ufunc_1dim,
        position="last",
    )
    return out


# TODO: Add window arg
# TODO: Inverse window arg to tolerate holes?
def run_bounds(mask: xr.DataArray, dim: str = "time", coord: bool | str = True):
    """
    Return the start and end dates of boolean runs along a dimension.

    Parameters
    ----------
    mask : xr.DataArray
        Boolean array.
    dim : str
        Dimension along which to look for runs.
    coord : bool or str
        If `True`, return values of the coordinate, if a string, returns values from `dim.dt.<coord>`.
        If `False`, return indexes.

    Returns
    -------
    xr.DataArray
        With ``dim`` reduced to "events" and "bounds". The events dim is as long as needed, padded with NaN or NaT.
    """
    if uses_dask(mask):
        raise NotImplementedError("Dask arrays not supported as we can't know the final event number before computing.")

    diff = xr.concat((mask.isel({dim: [0]}).astype(int), mask.astype(int).diff(dim)), dim)

    nstarts = (diff == 1).sum(dim).max().item()

    def _get_indices(arr, *, N):
        out = np.full((N,), np.nan, dtype=float)
        inds = np.where(arr)[0]
        out[: len(inds)] = inds
        return out

    starts = xr.apply_ufunc(
        _get_indices,
        diff == 1,
        input_core_dims=[[dim]],
        output_core_dims=[["events"]],
        kwargs={"N": nstarts},
        vectorize=True,
    )

    ends = xr.apply_ufunc(
        _get_indices,
        diff == -1,
        input_core_dims=[[dim]],
        output_core_dims=[["events"]],
        kwargs={"N": nstarts},
        vectorize=True,
    )

    if coord:
        crd = mask[dim]
        if isinstance(coord, str):
            crd = getattr(crd.dt, coord)

        starts = lazy_indexing(crd, starts)
        ends = lazy_indexing(crd, ends)
    return xr.concat((starts, ends), "bounds")


def keep_longest_run(da: xr.DataArray, dim: str = "time", freq: str | None = None) -> xr.DataArray:
    """
    Keep the longest run along a dimension.

    Parameters
    ----------
    da : xr.DataArray
        Boolean array.
    dim : str
        Dimension along which to check for the longest run.
    freq : str
        Resampling frequency.

    Returns
    -------
    xr.DataArray, [bool]
        Boolean array similar to da but with only one run, the (first) longest.
    """
    # Get run lengths
    rls = rle(da, dim)

    def _get_out(_rls):  # numpydoc ignore=GL08
        _out = xr.where(
            # Construct an integer array and find the max
            _rls[dim].copy(data=np.arange(_rls[dim].size)) == _rls.argmax(dim),
            _rls + 1,  # Add one to the First longest run
            _rls,
        )
        _out = _out.ffill(dim) == _out.max(dim)
        return _out

    if freq is not None:
        out = resample_map(rls, dim, freq, _get_out)
    else:
        out = _get_out(rls)

    return da.copy(data=out.transpose(*da.dims).data)


def runs_with_holes(
    da_start: xr.DataArray,
    window_start: int,
    da_stop: xr.DataArray,
    window_stop: int,
    dim: str = "time",
) -> xr.DataArray:
    """
    Extract events, i.e. runs whose starting and stopping points are defined through run length conditions.

    Parameters
    ----------
    da_start : xr.DataArray
        Input array where run sequences are searched to define the start points in the main runs.
    window_start : int
        Number of True (1) values needed to start a run in `da_start`.
    da_stop : xr.DataArray
        Input array where run sequences are searched to define the stop points in the main runs.
    window_stop : int
        Number of True (1) values needed to start a run in `da_stop`.
    dim : str
        Dimension name.

    Returns
    -------
    xr.DataArray
        Output array with 1's when in a run sequence and with 0's elsewhere.

    Notes
    -----
    A season (as defined in ``season``) could be considered as an event with ``window_stop == window_start``
    and ``da_stop == 1 - da_start``, although it has more constraints on when to start and stop a run through
    the `date` argument and only one season can be found.
    """
    da_start = da_start.astype(int).fillna(0)
    da_stop = da_stop.astype(int).fillna(0)

    start_runs = _cumsum_reset(da_start, dim=dim, index="first")
    stop_runs = _cumsum_reset(da_stop, dim=dim, index="first")
    start_positions = xr.where(start_runs >= window_start, 1, np.nan)
    stop_positions = xr.where(stop_runs >= window_stop, 0, np.nan)

    # start positions (1) are f-filled until a stop position (0) is met
    runs = stop_positions.combine_first(start_positions).ffill(dim=dim).fillna(0)
    return runs


def season_start(
    da: xr.DataArray,
    window: int,
    mid_date: DayOfYearStr | None = None,
    dim: str = "time",
    coord: str | bool | None = False,
) -> xr.DataArray:
    """
    Start of a season.

    See :py:func:`season`.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive values to start and end the season.
    mid_date : DayOfYearStr, optional
        The date (in MM-DD format) that a season must include to be considered valid.
    dim : str
        Dimension along which to calculate the season (default: 'time').
    coord : Optional[str]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').

    Returns
    -------
    xr.DataArray
        Start of the season, units depend on `coord`.

    See Also
    --------
    season : Calculate the bounds of a season along a dimension.
    season_end : End of a season.
    season_length : Length of a season.
    """
    return first_run_before_date(da, window=window, date=mid_date, dim=dim, coord=coord)


def season_end(
    da: xr.DataArray,
    window: int,
    mid_date: DayOfYearStr | None = None,
    dim: str = "time",
    coord: str | bool | None = False,
    _beg: xr.DataArray | None = None,
) -> xr.DataArray:
    """
    End of a season.

    See :py:func:`season`. Similar to :py:func:`first_run_after_date` but here a season
    must have a start for an end to be valid. Also, if no end is found but a start was found
    the end is set to the last element of the series.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive values to start and end the season.
    mid_date : DayOfYearStr, optional
        The date (in MM-DD format) that a run must include to be considered valid.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    coord : str, optional
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').
    _beg : xr.DataArray, optional
        If given, the start of the season. This is used to avoid recomputing the start.

    Returns
    -------
    xr.DataArray
        End of the season, units depend on `coord`.
        If there is a start is found but no end, the end is set to the last element.

    See Also
    --------
    season : Calculate the bounds of a season along a dimension.
    season_start : Start of a season.
    season_length : Length of a season.
    """
    # Fast path for `season` and `season_length`
    if _beg is not None:
        beg = _beg
    else:
        beg = season_start(da, window=window, dim=dim, mid_date=mid_date, coord=False)
    # Invert the condition and mask all values after beginning
    # we fillna(0) as so to differentiate series with no runs and all-nan series
    not_da = (~da).where(da[dim].copy(data=np.arange(da[dim].size)) >= beg.fillna(0))
    end = first_run_after_date(not_da, window=window, dim=dim, date=mid_date, coord=False)
    if _beg is None:
        # Where end is null and beg is not null (valid data, no end detected), put the last index
        # Don't do this in the fast path, so that the length can use this information
        end = xr.where(end.isnull() & beg.notnull(), da[dim].size - 1, end)
        end = end.where(beg.notnull())
    if coord:
        crd = da[dim]
        if isinstance(coord, str):
            crd = getattr(crd.dt, coord)
        end = lazy_indexing(crd, end)
    return end


def season(
    da: xr.DataArray,
    window: int,
    mid_date: DayOfYearStr | None = None,
    dim: str = "time",
    stat: str | None = None,
    coord: str | bool | None = False,
) -> xr.Dataset:
    """
    Calculate the bounds of a season along a dimension.

    A "season" is a run of True values that may include breaks under a given length (`window`).
    The start is computed as the first run of `window` True values, and the end as the first subsequent run
    of  `window` False values. The end element is the first element after the season.
    If a date is given, it must be included in the season
    (i.e. the start cannot occur later and the end cannot occur earlier).

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive values to start and end the season.
    mid_date : DayOfYearStr, optional
        The date (in MM-DD format) that a run must include to be considered valid.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    stat : str, optional
        Not currently implemented.
        If not None, return a statistic of the season. The statistic is calculated on the season's values.
    coord : Optional[str]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').

    Returns
    -------
    xr.Dataset
        The Dataset variables:
            start : start of the season (index or units depending on ``coord``)
            end : end of the season (index or units depending on ``coord``)
            length : length of the season (in number of elements along ``dim``)

    See Also
    --------
    season_start : Start of a season.
    season_end : End of a season.
    season_length : Length of a season.

    Notes
    -----
    The run can include holes of False or NaN values, so long as they do not exceed the window size.

    If a date is given, the season start and end are forced to be on each side of this date. This means that
    even if the "real" season has been over for a long time, this is the date used in the length calculation.
    e.g. Length of the "warm season", where T > 25°C, with date = 1st August. Let's say the temperature is over
    25 for all June, but July and august have very cold temperatures. Instead of returning 30 days (June),
    the function will return 61 days (July + June).

    The season's length is always the difference between the end and the start. Except if no season end was
    found before the end of the data. In that case the end is set to last element and the length is set to
    the data size minus the start index. Thus, for the specific case, :math:`length = end - start + 1`,
    because the end falls on the last element of the season instead of the subsequent one.
    """
    beg = season_start(da, window=window, dim=dim, mid_date=mid_date, coord=False)
    # Use fast path in season_end : no recomputing of start,
    # no masking of end where beg.isnull(), and don't set end if none found
    end = season_end(da, window=window, dim=dim, mid_date=mid_date, _beg=beg, coord=False)
    # Three cases :
    #           start     no start
    # end       e - s        0
    # no end   size - s      0
    # da is boolean, so we have no way of knowing if the absence of season
    # is due to missing values or to an actual absence.
    length = xr.where(
        beg.isnull(),
        0,
        # Where there is no end, from the start to the boundary
        xr.where(end.isnull(), da[dim].size - beg, end - beg),
    )
    # Now masks ends
    # Still give an end if we didn't find any : put the last element
    # This breaks the length = end - beg, but yields a truer length
    end = xr.where(end.isnull() & beg.notnull(), da[dim].size - 1, end)
    end = end.where(beg.notnull())

    if coord:
        crd = da[dim]
        if isinstance(coord, str):
            crd = getattr(crd.dt, coord)
            coordstr = coord
        else:
            coordstr = dim
        beg = lazy_indexing(crd, beg)
        end = lazy_indexing(crd, end)
    else:
        coordstr = "index"

    out = xr.Dataset({"start": beg, "end": end, "length": length})
    out.start.attrs.update(
        long_name="Start of the season.",
        description=f"First {coordstr} of a run of at least {window} steps respecting the condition.",
    )
    out.end.attrs.update(
        long_name="End of the season.",
        description=f"First {coordstr} of a run of at least {window} "
        "steps breaking the condition, starting after `start`.",
    )
    out.length.attrs.update(
        long_name="Length of the season.",
        description="Number of steps of the original series in the season, between 'start' and 'end'.",
    )
    return out


def season_length(
    da: xr.DataArray,
    window: int,
    mid_date: DayOfYearStr | None = None,
    dim: str = "time",
) -> xr.DataArray:
    """
    Length of a season.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive values to start and end the season.
    mid_date : DayOfYearStr, optional
        The date (in MM-DD format) that a run must include to be considered valid.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').

    Returns
    -------
    xr.DataArray, [int]
        Length of the season, in number of elements along dimension `time`.

    See Also
    --------
    season : Calculate the bounds of a season along a dimension.
    season_start : Start of a season.
    season_end : End of a season.
    """
    seas = season(da, window, mid_date, dim, coord=False)
    return seas.length


def run_end_after_date(
    da: xr.DataArray,
    window: int,
    date: DayOfYearStr = "07-01",
    dim: str = "time",
    coord: bool | str | None = "dayofyear",
) -> xr.DataArray:
    """
    Return the index of the first item after the end of a run after a given date.

    The run must begin before the date.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
    date : str
        The date after which to look for the end of a run.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    coord : Optional[Union[bool, str]]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of last item in last valid run.
        Returns np.nan if there are no valid runs.
    """
    mid_idx = index_of_date(da[dim], date, max_idxs=1, default=0)
    if mid_idx.size == 0:  # The date is not within the group. Happens at boundaries.
        return xr.full_like(da.isel({dim: 0}), np.nan, float).drop_vars(dim)

    end = first_run(
        (~da).where(da[dim] >= da[dim][mid_idx][0]),
        window=window,
        dim=dim,
        coord=coord,
    )
    beg = first_run(da.where(da[dim] < da[dim][mid_idx][0]), window=window, dim=dim)

    if coord:
        last = da[dim][-1]
        if isinstance(coord, str):
            last = getattr(last.dt, coord)
    else:
        last = da[dim].size - 1

    end = xr.where(end.isnull() & beg.notnull(), last, end)
    return end.where(beg.notnull()).drop_vars(dim, errors="ignore")


def first_run_after_date(
    da: xr.DataArray,
    window: int,
    date: DayOfYearStr | None = "07-01",
    dim: str = "time",
    coord: bool | str | None = "dayofyear",
) -> xr.DataArray:
    """
    Return the index of the first item of the first run after a given date.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
    date : DayOfYearStr
        The date after which to look for the run.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    coord : Optional[Union[bool, str]]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of first item in the first valid run.
        Returns np.nan if there are no valid runs.
    """
    mid_idx = index_of_date(da[dim], date, max_idxs=1, default=0)
    if mid_idx.size == 0:  # The date is not within the group. Happens at boundaries.
        return xr.full_like(da.isel({dim: 0}), np.nan, float).drop_vars(dim)

    return first_run(
        da.where(da[dim] >= da[dim][mid_idx][0]),
        window=window,
        dim=dim,
        coord=coord,
    )


def last_run_before_date(
    da: xr.DataArray,
    window: int,
    date: DayOfYearStr = "07-01",
    dim: str = "time",
    coord: bool | str | None = "dayofyear",
) -> xr.DataArray:
    """
    Return the index of the last item of the last run before a given date.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
    date : DayOfYearStr
        The date before which to look for the last event.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    coord : Optional[Union[bool, str]]
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (ex: 'dayofyear').

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of last item in last valid run.
        Returns np.nan if there are no valid runs.
    """
    mid_idx = index_of_date(da[dim], date, default=-1)

    if mid_idx.size == 0:  # The date is not within the group. Happens at boundaries.
        return xr.full_like(da.isel({dim: 0}), np.nan, float).drop_vars(dim)

    run = da.where(da[dim] <= da[dim][mid_idx][0])
    return last_run(run, window=window, dim=dim, coord=coord)


def first_run_before_date(
    da: xr.DataArray,
    window: int,
    date: DayOfYearStr | None = "07-01",
    dim: str = "time",
    coord: bool | str | None = "dayofyear",
) -> xr.DataArray:
    """
    Return the index of the first item of the first run before a given date.

    Parameters
    ----------
    da : xr.DataArray
        Input N-dimensional DataArray (boolean).
    window : int
        Minimum duration of consecutive run to accumulate values.
    date : DayOfYearStr
        The date before which to look for the run.
    dim : str
        Dimension along which to calculate consecutive run (default: 'time').
    coord : bool or str, optional
        If not False, the function returns values along `dim` instead of indexes.
        If `dim` has a datetime dtype, `coord` can also be a str of the name of the
        DateTimeAccessor object to use (e.g. 'dayofyear').

    Returns
    -------
    xr.DataArray
        Index (or coordinate if `coord` is not False) of first item in the first valid run.
        Returns np.nan if there are no valid runs.
    """
    if date is not None:
        mid_idx = index_of_date(da[dim], date, max_idxs=1, default=0)
        if mid_idx.size == 0:  # The date is not within the group. Happens at boundaries.
            return xr.full_like(da.isel({dim: 0}), np.nan, float).drop_vars(dim)
        # Mask anything after the mid_date + window - 1
        # Thus, the latest run possible can begin on the day just before mid_idx
        da = da.where(da[dim] < da[dim][mid_idx + window - 1][0])

    return first_run(
        da,
        window=window,
        dim=dim,
        coord=coord,
    )


@njit
def _rle_1d(ia) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    y = ia[1:] != ia[:-1]  # pairwise unequal (string safe)
    i = np.append(np.nonzero(y)[0], ia.size - 1)  # must include last element position
    rl = np.diff(np.append(-1, i))  # run lengths
    pos = np.cumsum(np.append(0, rl))[:-1]  # positions
    return ia[i], rl, pos


def rle_1d(
    arr: int | float | bool | Sequence[int | float | bool] | xr.DataArray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Return the length, starting position and value of consecutive identical values.

    In opposition to py:func:`rle`, this is an actuel run length encoder.

    Parameters
    ----------
    arr : int or float or bool or Sequence[Union[int, float, bool]] or xr.DataArray
        Array of values to be parsed.

    Returns
    -------
    values : np.ndarray
        The values taken by arr over each run.
    run_lengths : np.ndarray
        The length of each run.
    start_positions : np.ndarray
        The starting index of each run.

    Examples
    --------
    >>> from xclim.indices.run_length import rle_1d
    >>> a = [1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3]
    >>> rle_1d(a)
    RLE_1D(values=array([1, 2, 3]), run_lengths=array([2, 4, 6]), start_positions=array([0, 2, 6]))
    """
    ia = np.asarray(arr)
    n = len(ia)
    RLE_1D = namedtuple("RLE_1D", ["values", "run_lengths", "start_positions"])

    if n == 0:
        warn("run length array empty")
        # Returning None makes some other 1d func below fail.
        return RLE_1D(np.array(np.nan), np.array([0]), np.array(np.nan))
    return RLE_1D(*_rle_1d(ia))


def first_run_1d(arr: Sequence[int | float], window: int) -> int | float:
    """
    Return the index of the first item of a run of at least a given length.

    Parameters
    ----------
    arr : sequence of int or float
        Input array.
    window : int
        Minimum duration of consecutive run to accumulate values.

    Returns
    -------
    int or np.nan
        Index of first item in first valid run.
        Returns np.nan if there are no valid runs.
    """
    v, rl, pos = rle_1d(arr)
    ind = np.where(v * rl >= window, pos, np.inf).min()  # noqa

    if np.isinf(ind):
        return np.nan
    return ind


def statistics_run_1d(arr: Sequence[bool], reducer: str, window: int) -> int:
    """
    Return statistics on lengths of run of identical values.

    Parameters
    ----------
    arr : Sequence of bool
        Input array (bool).
    reducer : {"mean", "sum", "min", "max", "std", "count"}
        Reducing function name.
    window : int
        Minimal length of runs to be included in the statistics.

    Returns
    -------
    int
        Statistics on length of runs.
    """
    v, rl = rle_1d(arr)[:2]
    if not np.any(v) or np.all(v * rl < window):
        return 0
    if reducer == "count":
        return (v * rl >= window).sum()
    func = getattr(np, f"nan{reducer}")
    return func(np.where(v * rl >= window, rl, np.nan))


def windowed_run_count_1d(arr: Sequence[bool], window: int) -> int:
    """
    Return the number of consecutive true values in array for runs at least as long as given duration.

    Parameters
    ----------
    arr : Sequence of bool
        Input array (bool).
    window : int
        Minimum duration of consecutive run to accumulate values.

    Returns
    -------
    int
        Total number of true values part of a consecutive run at least `window` long.
    """
    v, rl = rle_1d(arr)[:2]
    return np.where(v * rl >= window, rl, 0).sum()


def windowed_run_events_1d(arr: Sequence[bool], window: int) -> xr.DataArray:
    """
    Return the number of runs of a minimum length.

    Parameters
    ----------
    arr : Sequence of bool
        Input array (bool).
    window : int
        Minimum run length.

    Returns
    -------
    xr.DataArray, [int]
        Number of distinct runs of a minimum length.
    """
    v, rl, _ = rle_1d(arr)
    return (v * rl >= window).sum()


def windowed_run_count_ufunc(x: xr.DataArray | Sequence[bool], window: int, dim: str) -> xr.DataArray:
    """
    Dask-parallel version of windowed_run_count_1d.

    The number of consecutive true values in array for runs at least as long as given duration.

    Parameters
    ----------
    x : xr.DataArray or sequence of bool
        Input array (bool).
    window : int
        Minimum duration of consecutive run to accumulate values.
    dim : str
        Dimension along which to calculate windowed run.

    Returns
    -------
    xr.DataArray
        A function operating along the time dimension of a dask-array.
    """
    return xr.apply_ufunc(
        windowed_run_count_1d,
        x,
        input_core_dims=[[dim]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[int],
        keep_attrs=True,
        kwargs={"window": window},
    )


def windowed_run_events_ufunc(x: xr.DataArray | Sequence[bool], window: int, dim: str) -> xr.DataArray:
    """
    Dask-parallel version of windowed_run_events_1d.

    The number of runs at least as long as given duration.

    Parameters
    ----------
    x : xr.DataArray or sequence of bool
        Input array (bool).
    window : int
        Minimum run length.
    dim : str
        Dimension along which to calculate windowed run.

    Returns
    -------
    xr.DataArray
        A function operating along the time dimension of a dask-array.
    """
    return xr.apply_ufunc(
        windowed_run_events_1d,
        x,
        input_core_dims=[[dim]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[int],
        keep_attrs=True,
        kwargs={"window": window},
    )


def statistics_run_ufunc(
    x: xr.DataArray | Sequence[bool],
    reducer: str,
    window: int,
    dim: str = "time",
) -> xr.DataArray:
    """
    Dask-parallel version of statistics_run_1d.

    The {reducer} number of consecutive true values in array.

    Parameters
    ----------
    x : Sequence of bool
        Input array (bool).
    reducer : {'min', 'max', 'mean', 'sum', 'std'}
        Reducing function name.
    window : int
        Minimal length of runs.
    dim : str
        The dimension along which the runs are found.

    Returns
    -------
    xr.DataArray
        A function operating along the time dimension of a dask-array.
    """
    return xr.apply_ufunc(
        statistics_run_1d,
        x,
        input_core_dims=[[dim]],
        kwargs={"reducer": reducer, "window": window},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
    )


def first_run_ufunc(
    x: xr.DataArray | Sequence[bool],
    window: int,
    dim: str,
) -> xr.DataArray:
    """
    Dask-parallel version of first_run_1d.

    The first entry in array of consecutive true values.

    Parameters
    ----------
    x : xr.DataArray or sequence of bool
        Input array (bool).
    window : int
        Minimum run length.
    dim : str
        The dimension along which the runs are found.

    Returns
    -------
    xr.DataArray
        A function operating along the time dimension of a dask-array.
    """
    ind = xr.apply_ufunc(
        first_run_1d,
        x,
        input_core_dims=[[dim]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=True,
        kwargs={"window": window},
    )

    return ind


def index_of_date(
    time: xr.DataArray,
    date: DateStr | DayOfYearStr | None,
    max_idxs: int | None = None,
    default: int = 0,
) -> np.ndarray:
    """
    Get the index of a date in a time array.

    Parameters
    ----------
    time : xr.DataArray
        An array of datetime values, any calendar.
    date : DayOfYearStr or DateStr, optional
        A string in the "yyyy-mm-dd" or "mm-dd" format.
        If None, returns default.
    max_idxs : int, optional
        Maximum number of returned indexes.
    default : int
        Index to return if date is None.

    Returns
    -------
    numpy.ndarray
        1D array of integers, indexes of `date` in `time`.

    Raises
    ------
    ValueError
        If there are most instances of `date` in `time` than `max_idxs`.
    """
    if date is None:
        return np.array([default])
    if len(date.split("-")) == 2:
        date = f"1840-{date}"
        date = datetime.strptime(date, "%Y-%m-%d")
        year_cond = True
    else:
        date = datetime.strptime(date, "%Y-%m-%d")
        year_cond = time.dt.year == date.year

    idxs = np.where(year_cond & (time.dt.month == date.month) & (time.dt.day == date.day))[0]
    if max_idxs is not None and idxs.size > max_idxs:
        raise ValueError(f"More than {max_idxs} instance of date {date} found in the coordinate array.")
    return idxs


def suspicious_run_1d(
    arr: np.ndarray,
    window: int = 10,
    op: Literal[">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"] = ">",
    thresh: float | None = None,
) -> np.ndarray:
    """
    Return `True` where the array contains a run of identical values.

    Parameters
    ----------
    arr : numpy.ndarray
        Array of values to be parsed.
    window : int
        Minimum run length.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Operator for threshold comparison. Defaults to ">".
    thresh : float, optional
        Threshold compared against which values are checked for identical values.

    Returns
    -------
    numpy.ndarray
        Whether the data points are part of a run of identical values or not.
    """
    v, rl, pos = rle_1d(arr)
    sus_runs = rl >= window
    if thresh is not None:
        if op in {">", "gt"}:
            sus_runs = sus_runs & (v > thresh)
        elif op in {"<", "lt"}:
            sus_runs = sus_runs & (v < thresh)
        elif op in {"==", "eq"}:
            sus_runs = sus_runs & (v == thresh)
        elif op in {"!=", "ne"}:
            sus_runs = sus_runs & (v != thresh)
        elif op in {">=", "gteq", "ge"}:
            sus_runs = sus_runs & (v >= thresh)
        elif op in {"<=", "lteq", "le"}:
            sus_runs = sus_runs & (v <= thresh)
        else:
            raise NotImplementedError(f"{op}")

    out = np.zeros_like(arr, dtype=bool)
    for st, l in zip(pos[sus_runs], rl[sus_runs], strict=False):  # noqa: E741
        out[st : st + l] = True  # noqa: E741
    return out


def suspicious_run(
    arr: xr.DataArray,
    dim: str = "time",
    window: int = 10,
    op: Literal[">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"] = ">",
    thresh: float | None = None,
) -> xr.DataArray:
    """
    Return `True` where the array contains has runs of identical values, vectorized version.

    In opposition to other run length functions, here the output has the same shape as the input.

    Parameters
    ----------
    arr : xr.DataArray
        Array of values to be parsed.
    dim : str
        Dimension along which to check for runs (default: "time").
    window : int
        Minimum run length.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Operator for threshold comparison, defaults to ">".
    thresh : float, optional
        Threshold above which values are checked for identical values.

    Returns
    -------
    xarray.DataArray
        A boolean array of the same shape as the input, indicating where runs of identical values are found.
    """
    return xr.apply_ufunc(
        suspicious_run_1d,
        arr,
        input_core_dims=[[dim]],
        output_core_dims=[[dim]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[bool],
        keep_attrs=True,
        kwargs={"window": window, "op": op, "thresh": thresh},
    )


def _find_events(da_start, da_stop, data, window_start, window_stop):
    """
    Actual finding of events for each period.

    Get basic blocks to work with, our runs with holes and the lengths of those runs.
    Series of ones indicating where we have continuous runs with pauses
    not exceeding `window_stop`
    """
    runs = runs_with_holes(da_start, window_start, da_stop, window_stop)

    # Compute the length of freezing rain events
    # I think int16 is safe enough, fillna first to suppress warning
    ds = rle(runs).fillna(np.int16(0)).to_dataset(name="event_length")
    # Time duration where the precipitation threshold is exceeded during an event
    # (duration of complete run - duration of holes in the run )
    ds["event_effective_length"] = _cumsum_reset_xr(
        da_start.where(runs == 1), dim="time", index="first", reset_on_zero=False
    ).astype(np.int16)

    if data is not None:
        # Ex: Cumulated precipitation in a given freezing rain event
        ds["event_sum"] = _cumsum_reset_xr(data.where(runs == 1), dim="time", index="first", reset_on_zero=False)

    # Keep time as a variable, it will be used to keep start of events
    ds["event_start"] = ds["time"].broadcast_like(ds)  # .astype(int)
    # We convert to an integer for the filtering, time object won't do well in the apply_ufunc/vectorize
    time_min = ds.event_start.min()
    ds["event_start"] = ds.event_start.copy(
        data=(ds.event_start - time_min).values.astype("timedelta64[s]").astype(int)
    )

    # Filter events: Reduce time dimension
    def _filter_events(da, rl, max_event_number):
        out = np.full(max_event_number, np.nan)
        events_start = da[rl > 0]
        out[: len(events_start)] = events_start
        return out

    # Dask inputs need to be told their length before computing anything.
    max_event_number = int(np.ceil(da_start.time.size / (window_start + window_stop)))
    ds = xr.apply_ufunc(
        _filter_events,
        ds,
        ds.event_length,
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[["event"]],
        kwargs={"max_event_number": max_event_number},
        dask_gufunc_kwargs={"output_sizes": {"event": max_event_number}},
        dask="parallelized",
        vectorize=True,
    )

    # convert back start to a time
    if time_min.dtype == "O":
        # Can't add a numpy array of timedeltas to an array of cftime (because they have non-compatible dtypes)
        # Also, we can't add cftime to NaTType. So we fill with negative timedeltas and mask them after the addition
        # Also, pd.to_timedelta can only handle 1D input, so we vectorize

        def _get_start_cftime(deltas, time_min=None):
            starts = time_min + pd.to_timedelta(deltas, "s").to_pytimedelta()
            starts[starts < time_min] = np.nan
            return starts

        ds["event_start"] = xr.apply_ufunc(
            _get_start_cftime,
            ds.event_start.fillna(-1),
            input_core_dims=[("event",)],
            output_core_dims=[("event",)],
            vectorize=True,
            dask="parallelized",
            kwargs={"time_min": time_min.item()},
            output_dtypes=[time_min.dtype],
        )
    else:
        ds["event_start"] = ds.event_start.copy(data=time_min.values + ds.event_start.data.astype("timedelta64[s]"))

    ds["event"] = np.arange(1, ds.event.size + 1)
    ds["event_length"].attrs["units"] = "1"
    ds["event_effective_length"].attrs["units"] = "1"
    ds["event_start"].attrs["units"] = ""
    if data is not None:
        ds["event_sum"].attrs["units"] = data.units
    return ds


# TODO: Implement more event stats ? (max, effective sum, etc)
def find_events(
    condition: xr.DataArray,
    window: int,
    condition_stop: xr.DataArray | None = None,
    window_stop: int = 1,
    data: xr.DataArray | None = None,
    freq: str | None = None,
) -> xr.Dataset:
    """
    Find events (runs).

    An event starts with a run of ``window`` consecutive True values in the condition
    and stops with ``window_stop`` consecutive True values in the stop condition.

    This returns a Dataset with each event along an `event` dimension.
    It does not perform statistics over the events like other function in this module do.

    Parameters
    ----------
    condition : DataArray of bool
        The boolean mask, true where the start condition of the event is fulfilled.
    window : int
        The number of consecutive True values for an event to start.
    condition_stop : DataArray of bool, optional
        The stopping boolean mask, true where the end condition of the event is fulfilled.
        Defaults to the opposite of `condition`.
    window_stop : int
        The number of consecutive True values in ``condition_stop`` for an event to end.
        Defaults to 1.
    data : DataArray, optional
        The actual data. If present, its sum within each event is added to the output.
    freq : str, optional
        A frequency to divide the data into periods. If absent, the output has not time dimension.
        If given, the events are searched within in each resample period independently.

    Returns
    -------
    xr.Dataset, same shape as the data (and the time dimension is resample or removed, according to ``freq``).
        The Dataset has the following variables:
            event_length: The number of time steps in each event
            event_effective_length: The number of time steps of even event where the start condition is true.
            event_start: The datetime of the start of the run.
            event_sum: The sum within each event, only considering steps where start condition is true (if ``data``).
    """
    if condition_stop is None:
        condition_stop = ~condition

    if freq is None:
        return _find_events(condition, condition_stop, data, window, window_stop)

    ds = xr.Dataset({"da_start": condition, "da_stop": condition_stop})
    if data is not None:
        ds = ds.assign(data=data)
    return ds.resample(time=freq).map(
        lambda grp: _find_events(grp.da_start, grp.da_stop, grp.get("data", None), window, window_stop)
    )



"""
Missing Values Identification
=============================

Indicators may use different criteria to determine whether a computed indicator value should be
considered missing. In some cases, the presence of any missing value in the input time series should result in a
missing indicator value for that period. In other cases, a minimum number of valid values or a percentage of missing
values should be enforced. The World Meteorological Organisation (WMO) suggests criteria based on the number of
consecutive and overall missing values per month.

`xclim` has a registry of missing value detection algorithms that can be extended by users to customize the behavior
of indicators. Once registered, algorithms can be used by setting the global option as
``xc.set_options(check_missing="method")`` or within indicators by setting the `missing` attribute of an
`Indicator` subclass. By default, `xclim` registers the following algorithms:

 * `any`: A result is missing if any input value is missing.
 * `at_least_n`: A result is missing if less than a given number of valid values are present.
 * `pct`: A result is missing if more than a given fraction of its values are missing.
 * `wmo`: A result is missing if 11 days are missing, or 5 consecutive values are missing in a month.

To define another missing value algorithm, subclass :py:class:`MissingBase` and decorate it with
:py:func:`xclim.core.options.register_missing_method`. See subclassing guidelines in ``MissingBase``'s doc.
"""

from __future__ import annotations

import textwrap

import numpy as np
import xarray as xr

from xclim.core.calendar import (
    compare_offsets,
    is_offset_divisor,
    parse_offset,
    select_time,
)
from xclim.core.options import (
    CHECK_MISSING,
    MISSING_METHODS,
    MISSING_OPTIONS,
    OPTIONS,
    register_missing_method,
)

__all__ = [
    "at_least_n_valid",
    "expected_count",
    "missing_any",
    "missing_from_context",
    "missing_pct",
    "missing_wmo",
    "register_missing_method",
]


# Mapping from sub-monthly CFtime freq strings to numpy timedelta64 units
# Only "minute" is different between the two
_freq_to_timedelta = {"min": "m"}


def expected_count(
    time: xr.DataArray,
    freq: str | None = None,
    src_timestep: str | None = None,
    **indexer,
) -> xr.DataArray:
    """
    Get expected number of step of length ``src_timestep`` per each resampling period
    ``freq`` that ``time`` covers.

    The determination of the resampling periods intersecting with the input array are
    done following xarray's and pandas' heuristics. The input coordinate needs not be
    continuous if `src_timestep` is given.

    Parameters
    ----------
    time : xr.DataArray, optional
        Input time coordinate from which the final resample time coordinate is guessed.
    freq : str, optional.
        Resampling frequency. If not given or None, the count for the full time range is returned.
    src_timestep : str, Optional
        The expected input frequency. If not given, it will be inferred from the input array.
    **indexer : Indexer
        Time attribute and values over which to subset the array. For example, use season='DJF' to select winter
        values, month=1 to select January, or month=[6,7,8] to select summer months.
        If not indexer is given, all values are considered.
        See :py:func:`xc.core.calendar.select_time`.

    Returns
    -------
    xr.DataArray
        Integer array at the resampling frequency with the number of expected elements in each period.
    """
    if src_timestep is None:
        src_timestep = xr.infer_freq(time)
        if src_timestep is None:
            raise ValueError("A src_timestep must be passed when it can't be inferred from the data.")

    if freq is not None and not is_offset_divisor(src_timestep, freq):
        raise NotImplementedError(
            "Missing checks not implemented for timeseries resampled to a frequency that is not "
            f"aligned with the source timestep. {src_timestep} is not a divisor of {freq}."
        )

    # Ensure a DataArray constructed like we expect
    time = xr.DataArray(time.values, dims=("time",), coords={"time": time.values}, name="time")

    if freq:
        # We only want the resulting time index, the actual resampling method is not important.
        resamp = time.resample(time=freq).count()
        resamp_time = resamp.indexes["time"]
        _, _, is_start, _ = parse_offset(freq)
        if is_start:
            start_time = resamp_time
            end_time = start_time.shift(1, freq=freq)
        else:
            end_time = resamp_time
            start_time = end_time.shift(-1, freq=freq)
    else:  # freq=None, means the whole timeseries as a single period
        i = time.indexes["time"]
        start_time = i[:1]
        end_time = i[-1:]

    # don't forget : W is converted to 7D
    mult, base, _, _ = parse_offset(src_timestep)
    if indexer or base in "YAQM":
        # Create a full synthetic time series and compare the number of days with the original series.
        t = xr.date_range(
            start_time[0],
            end_time[-1],
            freq=src_timestep,
            calendar=time.dt.calendar,
            use_cftime=(start_time.dtype == "O"),
        )

        sda = xr.DataArray(data=np.ones(len(t)), coords={"time": t}, dims=("time",))
        if "doy_bounds" in indexer:
            # This is the only case supported by select_time where DataArrays are supported
            # TODO: What happens when this new dim makes sda too large ? How do we involve dask here ?
            da_bnds = [bnd for bnd in indexer["doy_bounds"] if isinstance(bnd, xr.DataArray)]
            sda = xr.broadcast(sda, *da_bnds, exclude=("time",))[0]

        st = select_time(sda, **indexer)
        if freq:
            count = st.notnull().resample(time=freq).sum(dim="time")
        else:
            count = st.notnull().sum(dim="time")
    else:  # simpler way for sub monthly without indexer.
        delta = end_time - start_time
        unit = _freq_to_timedelta.get(base, base)
        n = delta.values.astype(f"timedelta64[{unit}]").astype(float) / mult

        if freq:
            count = xr.DataArray(n, coords={"time": resamp.time}, dims="time")
        else:
            count = xr.DataArray(n[0] + 1)
    return count


class MissingBase:
    r"""
    Base class used to determined where Indicator outputs should be masked.

    Subclasses should implement the ``is_missing``, ``validate`` and ``__init__``
    methods. The ``__init__`` is to be implemented in order to change the docstring
    and signature but is not expected to do anything other than the validation
    of the options, everything else should happen in the call (i.e. ``is_missing``).
    Subclasses can also override the ``_validate_src_timestep`` method to add restrictions
    on allowed values. That method should return False on invalid ``src_timestep``.

    Decorate subclasses with `xclim.core.options.register_missing_method` to add them
    to the registry before using them in an Indicator.
    """

    def __init__(self, **options):
        if not self.validate(**options):
            raise ValueError(f"Options {options} are not valid for {self.__class__.__name__}.")
        self.options = options

    @staticmethod
    def validate(**options):
        r"""
        Validate optional arguments.

        Parameters
        ----------
        **options : dict
            Optional arguments.

        Returns
        -------
        bool
            False if the options are not valid.
        """
        return True

    @staticmethod
    def is_valid(da: xr.DataArray, **indexer) -> xr.DataArray:
        r"""
        Return a boolean array indicating which values are valid.

        Parameters
        ----------
        da : xr.DataArray
            Input data.
        **indexer : {dim: indexer}, optional
            The time attribute and values over which to subset the array. For example, use season='DJF' to select winter
            values, month=1 to select January, or month=[6,7,8] to select summer months.
            See :py:func:`xclim.core.calendar.select_time`.

        Returns
        -------
        xr.DataArray
            Boolean array indicating which values are valid.
        """
        selected = select_time(da, **indexer)
        return selected.notnull()

    def _validate_src_timestep(self, src_timestep):
        return True

    def is_missing(
        self,
        valid: xr.DataArray,
        count: xr.DataArray,
        freq: str | None,
    ) -> xr.DataArray:
        """
        Return whether the values within each period should be considered missing or not.

        Must be implemented by subclasses.

        Parameters
        ----------
        valid : DataArray
            Boolean array of valid values (that has already been indexed).
        count : DataArray
            Indexer-aware integer array of number of expected elements at the resampling frequency.
        freq : str or None
            The resampling frequency, or None if the temporal dimension is to be collapsed.

        Returns
        -------
        DataArray
            Boolean array at the resampled frequency,
            True on the periods that should be considered missing.
        """
        raise NotImplementedError()

    def __call__(
        self,
        da: xr.DataArray,
        freq: str | None = None,
        src_timestep: str | None = None,
        **indexer,
    ) -> xr.DataArray:
        """
        Compute the missing period mask according to the object's algorithm.

        Parameters
        ----------
        da : xr.DataArray
            Input data, must have a "time" coordinate.
        freq : str, optional
            Resampling frequency. If None, a collapse of the temporal dimension is assumed.
        src_timestep : str, optional
            The expected source input frequency. If not given, it will be inferred from the input array.
        **indexer : Indexer
            Time attribute and values over which to subset the array. For example, use season='DJF' to select winter
            values, month=1 to select January, or month=[6,7,8] to select summer months.
            If not indexer is given, all values are considered.
            See :py:func:`xclim.core.calendar.select_time`.

        Returns
        -------
        DataArray
            Boolean array at the resampled frequency,
            True on the periods that should be considered missing or invalid.
        """
        if src_timestep is None:
            src_timestep = xr.infer_freq(da.time)
            if src_timestep is None:
                raise ValueError(
                    "The source timestep can't be inferred from the data, but it is required"
                    " to compute the missing values mask."
                )

        if not self._validate_src_timestep(src_timestep):
            raise ValueError(
                f"Input source timestep {src_timestep} is invalid for missing method {self.__class__.__name__}."
            )

        count = expected_count(da.time, freq=freq, src_timestep=src_timestep, **indexer)
        valid = self.is_valid(da, **indexer)
        return self.is_missing(valid, count, freq)

    def __repr__(self):
        opt_str = ", ".join([f"{k}={v}" for k, v in self.options.items()])
        return f"<{self.__class__.__name__}({opt_str})>"


# -----------------------------------------------
# --- Missing value identification algorithms ---
# -----------------------------------------------


@register_missing_method("any")
class MissingAny(MissingBase):
    """Mask periods as missing if any of its elements is missing or invalid."""

    def __init__(self):
        """Create a MissingAny object."""
        super().__init__()

    def is_missing(self, valid: xr.DataArray, count: xr.DataArray, freq: str | None) -> xr.DataArray:
        if freq is not None:
            valid = valid.resample(time=freq)
        # The number of valid values should fit the expected count.
        return valid.sum(dim="time") != count


# TODO: Make coarser method controllable.
class MissingTwoSteps(MissingBase):
    r"""
    Base class used to determine where Indicator outputs should be masked in a two-step process.

    In addition to what :py:class:`MissingBase` does, subclasses first perform the mask
    determination at some frequency and then resample at the (coarser) target frequency.
    This allows the application of specific methods at a finer resolution than the target one.
    The sub-groups are merged using the "Any" method : a group is invalid if any of its
    sub-groups are invalid.

    The first resampling frequency should be implemented as an additional "subfreq" option.
    A value of None means that only one resampling is done at the request target frequency.
    """

    def __call__(
        self,
        da: xr.DataArray,
        freq: str | None = None,
        src_timestep: str | None = None,
        **indexer,
    ) -> xr.DataArray:
        """
        Compute the missing period mask according to the object's algorithm.

        Parameters
        ----------
        da : xr.DataArray
            Input data, must have a "time" coordinate.
        freq : str, optional
            Target resampling frequency. If None, a collapse of the temporal dimension is assumed.
        src_timestep : str, optional
            The expected source input frequency. If not given, it will be inferred from the input array.
        **indexer : Indexer
            Time attribute and values over which to subset the array. For example, use season='DJF' to select winter
            values, month=1 to select January, or month=[6,7,8] to select summer months.
            If no indexer is given, all values are considered.
            See :py:func:`xclim.core.calendar.select_time`.

        Returns
        -------
        DataArray
            Boolean array at the resampled frequency,
            True on the periods that should be considered missing or invalid.
        """
        subfreq = self.options["subfreq"] or freq
        if subfreq is not None and freq is not None and compare_offsets(freq, "<", subfreq):
            raise ValueError(
                "The target resampling frequency cannot be finer than the first-step "
                f"frequency. Got : {subfreq} > {freq}."
            )
        miss = super().__call__(da, freq=subfreq, src_timestep=src_timestep, **indexer)
        if subfreq != freq:
            miss = MissingAny()(miss.where(~miss), freq, src_timestep=subfreq, **indexer)
        return miss


@register_missing_method("wmo")
class MissingWMO(MissingTwoSteps):
    """
    Mask periods as missing using the WMO criteria for missing days.

    The World Meteorological Organisation recommends that where monthly means are computed from daily values,
    it should be considered missing if either of these two criteria are met:

      – observations are missing for 11 or more days during the month;
      – observations are missing for a period of 5 or more consecutive days during the month.

    Stricter criteria are sometimes used in practice, with a tolerance of 5 missing values or 3 consecutive missing
    values.

    Notes
    -----
    If used at frequencies larger than a month, for example on an annual or seasonal basis, the function will return
    True if any month within a period is masked.
    """

    def __init__(self, nm: int = 11, nc: int = 5):
        """
        Create a MissingWMO object.

        Parameters
        ----------
        nm : int
            Minimal number of missing elements for a month to be masked.
        nc : int
            Minimal number of consecutive missing elements for a month to be masked.
        """
        super().__init__(nm=nm, nc=nc, subfreq="MS")

    @staticmethod
    def validate(nm: int, nc: int, subfreq: str | None = None):
        return nm < 31 and nc < 31

    def _validate_src_timestep(self, src_timestep):
        return src_timestep == "D"

    def is_missing(self, valid: xr.DataArray, count: xr.DataArray, freq: str) -> xr.DataArray:
        from xclim.indices import run_length as rl
        from xclim.indices.helpers import resample_map

        validr = valid.resample(time=freq)

        # Total number of missing or invalid days
        missing_days = count - validr.sum(dim="time")
        # Check if more than threshold is missing
        cond1 = missing_days >= self.options["nm"]

        # Check for consecutive invalid values
        # FIXME: This does not take holes in consideration
        longest_run = resample_map(~valid, "time", freq, rl.longest_run, map_blocks=True)
        cond2 = longest_run >= self.options["nc"]

        return cond1 | cond2


@register_missing_method("pct")
class MissingPct(MissingTwoSteps):
    """Mask periods as missing when there are more than a given percentage of missing days."""

    def __init__(self, tolerance: float = 0.1, subfreq: str | None = None):
        """
        Create a MissingPct object.

        Parameters
        ----------
        tolerance: float
            The maximum tolerated proportion of missing values,
            given as a number between 0 and 1.
        subfreq : str, optional
            If given, computes a mask at this frequency using this method and
            then resample at the target frequency using the "any" method on subgroups.
        """
        super().__init__(tolerance=tolerance, subfreq=subfreq)

    @staticmethod
    def validate(tolerance: float, subfreq: str | None = None):
        return 0 <= tolerance <= 1

    def is_missing(self, valid: xr.DataArray, count: xr.DataArray, freq: str | None) -> xr.DataArray:
        if freq is not None:
            valid = valid.resample(time=freq)

        # Total number of missing or invalid days
        missing_days = (count - valid.sum(dim="time")).fillna(count)
        return (missing_days / count) >= self.options["tolerance"]


@register_missing_method("at_least_n")
class AtLeastNValid(MissingTwoSteps):
    r"""
    Mask periods as missing if they don't have at least a given number of valid values.

    Ignores the expected count of elements.
    """

    def __init__(self, n: int = 20, subfreq: str | None = None):
        """
        Create an AtLeastNValid object.

        Parameters
        ----------
        n: float
            The minimum number of valid values needed.
        subfreq : str, optional
            If given, computes a mask at this frequency using this method and
            then resample at the target frequency using the "any" method on subgroups.
        """
        super().__init__(n=n, subfreq=subfreq)

    @staticmethod
    def validate(n: int, subfreq: str | None = None):
        return n > 0

    def is_missing(self, valid: xr.DataArray, count: xr.DataArray, freq: str | None) -> xr.DataArray:
        if freq is not None:
            valid = valid.resample(time=freq)
        nvalid = valid.sum(dim="time")
        return nvalid < self.options["n"]


# --------------------------
# --- Shortcut functions ---
# --------------------------
# These stand-alone functions hide the fact the algorithms are implemented in a class and make their use more
# user-friendly. This can also be useful for testing.


def missing_any(  # noqa: D103 # numpydoc ignore=GL08
    da: xr.DataArray, freq: str, src_timestep: str | None = None, **indexer
) -> xr.DataArray:
    """Return whether there are missing days in the array."""
    return MissingAny()(da, freq, src_timestep, **indexer)


def missing_wmo(  # noqa: D103 # numpydoc ignore=GL08
    da: xr.DataArray,
    freq: str,
    src_timestep: str | None = None,
    nm: int = 11,
    nc: int = 5,
    **indexer,
) -> xr.DataArray:
    return MissingWMO(nm=nm, nc=nc)(da, freq, src_timestep, **indexer)


def missing_pct(  # noqa: D103 # numpydoc ignore=GL08
    da: xr.DataArray,
    freq: str,
    src_timestep: str | None = None,
    tolerance: float = 0.1,
    subfreq: str | None = None,
    **indexer,
) -> xr.DataArray:
    return MissingPct(tolerance=tolerance, subfreq=subfreq)(da, freq, src_timestep, **indexer)


def at_least_n_valid(  # noqa: D103 # numpydoc ignore=GL08
    da: xr.DataArray,
    freq: str,
    src_timestep: str | None = None,
    n: int = 20,
    subfreq: str | None = None,
    **indexer,
) -> xr.DataArray:
    return AtLeastNValid(n=n, subfreq=subfreq)(da, freq, src_timestep, **indexer)


def missing_from_context(da: xr.DataArray, freq: str, src_timestep: str | None = None, **indexer) -> xr.DataArray:
    """
    Mask periods as missing according to the algorithm and options set in xclim's global options.

    The options can be manipulated with :py:func:`xclim.core.options.set_options`.

    Parameters
    ----------
    da : xr.DataArray
        Input data, must have a "time" coordinate.
    freq : str, optional
        Resampling frequency. If absent, a collapse of the temporal dimension is assumed.
    src_timestep : str, optional
        The expected source input frequency. If not given, it will be inferred from the input array.
    **indexer : Indexer
        Time attribute and values over which to subset the array. For example, use season='DJF' to select winter
        values, month=1 to select January, or month=[6,7,8] to select summer months.
        If not indexer is given, all values are considered.
        See :py:func:`xclim.core.calendar.select_time`.

    Returns
    -------
    DataArray
        Boolean array at the resampled frequency,
        True on the periods that should be considered missing or invalid.
    """
    method = OPTIONS[CHECK_MISSING]
    MissCls = MISSING_METHODS[method]
    opts = OPTIONS[MISSING_OPTIONS].get(method, {})
    return MissCls(**opts)(da, freq, src_timestep, **indexer)


def _get_convenient_doc(cls):
    maindoc = textwrap.dedent(cls.__doc__)
    initdoc = textwrap.dedent(cls.__init__.__doc__)
    calldoc = textwrap.dedent(cls.__call__.__doc__)

    params = []
    ip = 10000
    for i, line in enumerate(initdoc.split("\n")):
        if line.strip() == "Parameters":
            ip = i
        if i >= ip + 2 and line.strip():
            params.append(line)

    doc = [maindoc]
    if "\n" not in maindoc:
        doc.append("")

    ip = 10000
    for i, line in enumerate(calldoc.split("\n")):
        if line.strip() == "Parameters":
            ip = i
        elif "**indexer" in line:
            doc.extend(params)
        if i >= ip:
            doc.append(line)
    return "\n".join(doc)


missing_any.__doc__ = _get_convenient_doc(MissingAny)
missing_wmo.__doc__ = _get_convenient_doc(MissingWMO)
missing_pct.__doc__ = _get_convenient_doc(MissingPct)
at_least_n_valid.__doc__ = _get_convenient_doc(AtLeastNValid)



"""
CF-Convention Checking
======================

Utilities designed to verify the compliance of metadata with the CF-Convention.
"""

from __future__ import annotations

import fnmatch
import re
from collections.abc import Sequence

import xarray as xr

from xclim.core._exceptions import ValidationError
from xclim.core._types import VARIABLES
from xclim.core.options import cfcheck


@cfcheck
def check_valid(var: xr.DataArray, key: str, expected: str | Sequence[str]):
    r"""
    Check that a variable's attribute has one of the expected values and raise a ValidationError if otherwise.

    Parameters
    ----------
    var : xr.DataArray
        The variable to check.
    key : str
        The attribute to check.
    expected : str or sequence of str
        The expected value(s).

    Raises
    ------
    ValidationError
        If the attribute is not present or does not match the expected value(s).
    """
    att = getattr(var, key, None)
    if att is None:
        raise ValidationError(f"Variable does not have a `{key}` attribute.")
    if isinstance(expected, str):
        expected = [expected]
    for exp in expected:
        if fnmatch.fnmatch(att, exp):
            break
    else:
        raise ValidationError(
            f"Variable has a non-conforming {key}: Got `{att}`, expected `{expected}`",
        )


def cfcheck_from_name(varname: str, vardata: xr.DataArray, attrs: list[str] | None = None):
    """
    Perform cfchecks on a DataArray using specifications from xclim's default variables.

    Parameters
    ----------
    varname : str
        The name of the variable to check.
    vardata : xr.DataArray
        The variable to check.
    attrs : list of str, optional
        Attributes to check. Default is ["cell_methods", "standard_name"].

    Raises
    ------
    ValidationError
        If the variable does not meet the expected CF-Convention.
    """
    if attrs is None:
        attrs = ["cell_methods", "standard_name"]

    data = VARIABLES[varname]
    if "cell_methods" in data and "cell_methods" in attrs:
        _check_cell_methods(getattr(vardata, "cell_methods", None), data["cell_methods"])
    if "standard_name" in data and "standard_name" in attrs:
        check_valid(vardata, "standard_name", data["standard_name"])


@cfcheck
def _check_cell_methods(data_cell_methods: str, expected_method: str) -> None:
    if data_cell_methods is None:
        raise ValidationError("Variable does not have a `cell_methods` attribute.")
    EXTRACT_CELL_METHOD_REGEX = r"(\s*\S+\s*:(\s+[\w()-]+)+)(?!\S*:)"
    for m in re.compile(EXTRACT_CELL_METHOD_REGEX).findall(data_cell_methods):
        if expected_method in m[0]:
            return None
    raise ValidationError(
        f"Variable has a non-conforming cell_methods: "
        f"Got `{data_cell_methods}`, which do not include the expected "
        f"`{expected_method}`."
    )



from __future__ import annotations

import numpy as np
import pytest

from xclim import indices as xci
from xclim import land
from xclim.core.units import convert_units_to


class TestBaseFlowIndex:
    def test_simple(self, q_series):
        a = np.zeros(365) + 10
        a[10:17] = 1
        q = q_series(a)
        out = xci.base_flow_index(q)
        np.testing.assert_array_equal(out, 1.0 / a.mean())


class TestRBIndex:
    def test_simple(self, q_series):
        a = np.zeros(365)
        a[10] = 10
        q = q_series(a)
        out = xci.rb_flashiness_index(q)
        np.testing.assert_array_equal(out, 2)


class TestStandardizedStreamflow:
    @pytest.mark.slow
    def test_3d_data_with_nans(self, open_dataset):
        # test with data
        ds = open_dataset("Raven/q_sim.nc")
        q = ds.q_obs.sel(time=slice("2008")).rename("q")
        qMM = convert_units_to(q, "mm**3/s", context="hydro")  # noqa
        # put a nan somewhere
        qMM.values[10] = np.nan
        q.values[10] = np.nan

        out1 = land.standardized_streamflow_index(
            q,
            freq="MS",
            window=1,
            dist="genextreme",
            method="APP",
            fitkwargs={"floc": 0},
        )
        out2 = land.standardized_streamflow_index(
            qMM,
            freq="MS",
            window=1,
            dist="genextreme",
            method="APP",
            fitkwargs={"floc": 0},
        )
        np.testing.assert_array_almost_equal(out1, out2, 3)

    @pytest.mark.slow
    def test_3d_data_with_nans_value(self, open_dataset):
        # test with data
        ds = open_dataset("Raven/q_sim.nc")
        q = ds.q_obs.sel(time=slice("2008", "2018")).rename("q")
        q[{"time": 10}] = np.nan

        out1 = land.standardized_streamflow_index(
            q,
            freq=None,
            window=1,
            dist="genextreme",
            method="APP",
            fitkwargs={"floc": 0},
        )
        assert np.isnan(out1[{"time": 10}])


class TestSnwMax:
    def test_simple(self, snw_series):
        a = np.zeros(366)
        a[10:20] = np.arange(0, 10)
        snw = snw_series(a, start="1999-01-01")
        out = xci.snw_max(snw, freq="YS")
        np.testing.assert_array_equal(out, [9, 0])
        assert out.units == "kg m-2"


class TestSnwMaxDoy:
    def test_simple(self, snw_series):
        a = np.zeros(366)
        a[10] = 10
        snw = snw_series(a, start="1999-01-01")
        out = xci.snw_max_doy(snw, freq="YS")
        np.testing.assert_array_equal(out, [11, np.nan])
        assert out.attrs["units"] == "1"


class TestSnowMeltWEMax:
    def test_simple(self, snw_series):
        a = np.zeros(365)
        a[10:20] = np.arange(0, 10)
        a[20:25] = np.arange(10, 0, -2)
        snw = snw_series(a, start="1999-07-01")
        out = xci.snow_melt_we_max(snw)
        np.testing.assert_array_equal(out, 6)
        assert out.units == "kg m-2"


class TestMeltandPrecipMax:
    def test_simple(self, snw_series, pr_series):
        a = np.zeros(365)

        # 1 km / m2 of melt on day 11.
        a[10] = 1
        snw = snw_series(a, start="1999-07-01")

        # 1 kg/ m2 /d of rain on day 11
        b = np.zeros(365)
        b[11] = 1.0 / 60**2 / 24
        pr = pr_series(b, start="1999-07-01")

        out = xci.melt_and_precip_max(snw, pr)
        np.testing.assert_array_equal(out, 2)
        assert out.units == "kg m-2"


class TestFlowindex:
    def test_simple(self, q_series):
        a = np.ones(365 * 2) * 10
        a[10:50] = 50
        q = q_series(a)
        out = xci.flow_index(q, 0.95)
        np.testing.assert_array_equal(out, 5)


class TestHighflowfrequency:
    def test_simple(self, q_series):
        a = np.zeros(365 * 2)
        a[50:60] = 10
        a[200:210] = 20
        q = q_series(a)
        out = xci.high_flow_frequency(q, 9, freq="YS")
        np.testing.assert_array_equal(out, [20, 0])


class TestLowflowfrequency:
    def test_simple(self, q_series):
        a = np.ones(365 * 2) * 10
        a[50:60] = 1
        a[200:210] = 1
        q = q_series(a)
        out = xci.low_flow_frequency(q, 0.2, freq="YS")

        np.testing.assert_array_equal(out, [20, 0])


class TestAntecedentPrecipitationIndex:
    def test_simple(self, pr_series):
        a = np.ones(50) * 10
        a[15:20] = 20
        a[35:40] = 0
        pr = pr_series(a, units="mm d-1")
        out = xci.antecedent_precipitation_index(pr)
        np.testing.assert_allclose(out.max(), [101.65], atol=1e-2)
        np.testing.assert_allclose(out.min(), [13.83], atol=1e-2)

    def test_nan_present(self, pr_series):
        a = np.ones(50) * 10
        a[25] = np.nan
        pr = pr_series(a, units="mm d-1")
        window = 7
        out = xci.antecedent_precipitation_index(pr, window=window, p_exp=0.935)
        np.testing.assert_array_equal(out[25], [np.nan])

    def test_nan_start_window(self, pr_series):
        a = np.ones(50) * 10
        pr = pr_series(a, units="mm d-1")
        window = 7
        out = xci.antecedent_precipitation_index(pr, window=window, p_exp=0.935)
        np.testing.assert_array_equal(out[: window - 1], np.nan)

    def test_manual_calc(self, pr_series):
        a = np.ones(10) * 10
        pr = pr_series(a, units="mm d-1")
        window = 7
        p_exp = 0.935
        out = xci.antecedent_precipitation_index(pr, window=window, p_exp=p_exp)

        out_manual = np.zeros(out.shape) * np.nan
        for idx in range(pr.shape[0] - window + 1):
            idxend = window + idx
            weights = list(reversed([p_exp ** (ii + 1 - 1) for ii in range(window)]))
            weighted_sum = (pr[idx:idxend] * weights).sum()
            out_manual[idxend - 1] = weighted_sum
        np.testing.assert_allclose(out, out_manual, atol=1e-7)



"""
Conversion Indicators
=====================

This submodule contains indicators that converts CF-compliant variables from one to another.
For example, converting wind speed in cardinal directions to a vector magnitude and direction,
or converting snow depth to snow water equivalent. It also includes indicators that approximate
variables from multiple variables, such as calculating the mean temperature from daily maximum
and minimum temperatures.
"""

from __future__ import annotations

from ._conversion import *



"""Conversion and approximation functions."""

import warnings

from xclim.indices.converters import *

warnings.warn(
    "The `xclim.indices._conversion` submodule is deprecated and will be removed in a future release. "
    "Functions available here have been migrated to `xclim.indices.converters`.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = [
    "clausius_clapeyron_scaled_precipitation",
    "clearness_index",
    "dewpoint_from_specific_humidity",
    "fao_allen98",
    "heat_index",
    "humidex",
    "longwave_upwelling_radiation_from_net_downwelling",
    "mean_radiant_temperature",
    "potential_evapotranspiration",
    "prsn_to_prsnd",
    "prsnd_to_prsn",
    "rain_approximation",
    "relative_humidity",
    "saturation_vapor_pressure",
    "sfcwind_to_uas_vas",
    "shortwave_downwelling_radiation_from_clearness_index",
    "shortwave_upwelling_radiation_from_net_downwelling",
    "snd_to_snw",
    "snowfall_approximation",
    "snw_to_snd",
    "specific_humidity",
    "specific_humidity_from_dewpoint",
    "tas",
    "uas_vas_to_sfcwind",
    "universal_thermal_climate_index",
    "vapor_pressure",
    "vapor_pressure_deficit",
    "wind_chill_index",
    "wind_power_potential",
    "wind_profile",
]
